import requests
from bs4 import BeautifulSoup
from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template

NIMBLE_TOKEN = "YOUR_NIMBLE_API_KEY"


def nimble_google_search(query: str):
    """
    Use Nimble to get information about the query using Google Search.
    """
    url = "https://api.webit.live/api/v1/realtime/serp"
    headers = {
        "Authorization": f"Basic {NIMBLE_TOKEN}",
        "Content-Type": "application/json",
    }
    search_data = {
        "parse": True,
        "query": query,
        "search_engine": "google_search",
        "format": "json",
        "render": True,
        "country": "US",
        "locale": "en",
    }
    response = requests.get(url, json=search_data, headers=headers)
    data = response.json()
    results = data["parsing"]["entities"]["OrganicResult"]
    urls = [result.get("url", "") for result in results]
    search_results = {}
    for url in urls:
        content = get_content(url)
        search_results[url] = content
    return search_results


def get_content(url: str):
    data = []
    response = requests.get(url)
    content = response.content
    soup = BeautifulSoup(content, "html.parser")
    paragraphs = soup.find_all("p")
    for paragraph in paragraphs:
        data.append(paragraph.text)
    return "\n".join(data)


@openai.call(
    model="gpt-4o-mini",
    tools=[nimble_google_search],
    call_params={"tool_choice": "required"},
)
@prompt_template(
    """
    SYSTEM:
    You are a an expert at finding information on the web.
    Use the `nimble_google_search` function to find information on the web.
    Rewrite the question as needed to better find information on the web.

    USER:
    {question}
    """
)
def search(question: str): ...


class SearchResponse(BaseModel):
    sources: list[str] = Field(description="The sources of the results")
    answer: str = Field(description="The answer to the question")


@openai.call(model="gpt-4o-mini", response_model=list[SearchResponse])
@prompt_template(
    """
    SYSTEM:
    Extract the question, results, and sources to answer the question based on the results.

    Results:
    {results}

    USER:
    {question}
    """
)
def extract(question: str, results: str): ...


def run(question: str):
    response = search(question)
    if tool := response.tool:
        output = tool.call()
        result = extract(question, output)
        return result


print(run("What is the average price of a house in the United States?"))



"""Tests the `_utils.extract_tool_return` module."""

from pydantic import BaseModel

from mirascope.core.base._utils._extract_tool_return import extract_tool_return


def test_extract_tool_return() -> None:
    """Tests the `extract_tool_return` function."""

    class Book(BaseModel):
        title: str
        author: str

    book = extract_tool_return(
        Book,
        '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
        allow_partial=False,
    )
    assert isinstance(book, Book)
    book = extract_tool_return(Book, '{"title": "The Name', allow_partial=True)
    assert isinstance(book, Book)
    assert book.title == "The Name"
    assert book.author is None


def test_extract_tool_return_base_type() -> None:
    """Tests the `extract_tool_return` function with a base type."""

    title = extract_tool_return(
        str, '{"value": "The Name of the Wind"}', allow_partial=False
    )
    assert isinstance(title, str)
    assert title == "The Name of the Wind"
    title = extract_tool_return(str, '{"value": "The Name', allow_partial=True)
    assert isinstance(title, str)
    assert title == "The Name"



"""This module contains the setup_call function for Groq tools."""

import inspect
from collections.abc import Awaitable, Callable
from typing import Any, cast

from groq import AsyncGroq, Groq
from groq.types.chat import ChatCompletion, ChatCompletionMessageParam

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import GroqCallParams
from ..dynamic_config import GroqDynamicConfig
from ..tool import GroqTool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: Groq | AsyncGroq | None,
    fn: Callable[..., GroqDynamicConfig | Awaitable[GroqDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GroqDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GroqCallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
    str,
    list[ChatCompletionMessageParam],
    list[type[GroqTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, GroqTool, call_params
    )
    messages = cast(list[BaseMessageParam | ChatCompletionMessageParam], messages)
    messages = convert_message_params(messages)
    if json_mode:
        call_kwargs["response_format"] = {"type": "json_object"}
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        if messages[-1]["role"] == "user":
            if isinstance(messages[-1]["content"], str):
                messages[-1]["content"] += json_mode_content
            else:
                messages[-1]["content"] = list(messages[-1]["content"]) + [
                    {"type": "text", "text": json_mode_content.strip()}
                ]
        else:
            messages.append({"role": "user", "content": json_mode_content.strip()})
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = {
            "type": "function",
            "function": {"name": tool_types[0]._name()},
        }
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = AsyncGroq() if inspect.iscoroutinefunction(fn) else Groq()
    create = client.chat.completions.create

    return create, prompt_template, messages, tool_types, call_kwargs



"""The `groq_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import GroqCallParams
from .call_response import GroqCallResponse
from .call_response_chunk import GroqCallResponseChunk
from .dynamic_config import GroqDynamicConfig
from .stream import GroqStream
from .tool import GroqTool

groq_call = call_factory(
    TCallResponse=GroqCallResponse,
    TCallResponseChunk=GroqCallResponseChunk,
    TDynamicConfig=GroqDynamicConfig,
    TToolType=GroqTool,
    TStream=GroqStream,
    TCallParams=GroqCallParams,
    default_call_params=GroqCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
"""A decorator for calling the Groq API with a typed function.

usage docs: learn/calls.md

This decorator is used to wrap a typed function that calls the Groq API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.groq import groq_call


@groq_call("llama-3.1-8b-instant")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str):
    ...

response = recommend_book("fantasy")
print(response.content)
```

Args:
    model (str): The Groq model to use in the API call.
    stream (bool): Whether to stream the response from the API call.
    tools (list[BaseTool | Callable]): The tools to use in the Groq API call.
    response_model (BaseModel | BaseType): The response model into which the response
        should be structured.
    output_parser (Callable[[CohereCallResponse | ResponseModelT], Any]): A function for
        parsing the call response whose value will be returned in place of the original
        call response.
    json_mode (bool): Whether to use JSON Mode.
    client (object): An optional custom client to use in place of the default client.
    call_params (GroqCallParams): The `GroqCallParams` call parameters to use in the API
        call.

Returns:
    decorator (Callable): The decorator for turning a typed function into a Groq API
        call.
"""



from mirascope.core import BasePrompt, prompt_template


@prompt_template("Recommend a book cheaper than ${price:.2f}")
class BookRecommendationPrompt(BasePrompt):
    price: float


prompt = BookRecommendationPrompt(price=12.3456)
print(prompt)
# > Recommend a book cheaper than $12.34



import asyncio

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini")
@prompt_template("Summarize the plot of a {genre} movie")
async def summarize_movie(genre: str): ...


async def main():
    genres = ["action", "comedy", "drama", "sci-fi"]
    tasks = [summarize_movie(genre) for genre in genres]
    results = await asyncio.gather(*tasks)

    for genre, result in zip(genres, results, strict=False):
        print(f"{genre.capitalize()} movie summary:")
        print(result.content)
        print()


asyncio.run(main())



from mirascope.core import BasePrompt, metadata, prompt_template


@metadata({"tags": {"version:0001", "category:books"}})
@prompt_template("Recommend a {genre} book for a {age_group} reader.")
class BookRecommendationPrompt(BasePrompt):
    genre: str
    age_group: str


prompt = BookRecommendationPrompt(genre="fantasy", age_group="adult")
print(prompt.dump()["metadata"])
# > {'tags': {'version:0001', 'category:books'}}



import requests
from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel

from mirascope.core import openai, prompt_template

NIMBLE_TOKEN = "YOUR_NIMBLE_TOKEN"


def amazon_search(question: str):
    """Use Nimble to search for products on Amazon."""
    url = "https://api.webit.live/api/v1/realtime/web"
    headers = {
        "Authorization": f"Basic {NIMBLE_TOKEN}",
        "Content-Type": "application/json",
    }
    data = {
        "parse": True,
        "vendor": "amazon",
        "url": f"https://www.amazon.com/s?k={question}",
        "format": "json",
        "render": True,
        "country": "US",
        "locale": "en",
    }

    response = requests.post(url, headers=headers, json=data)
    json_response = response.json()
    products = [
        {
            "product_name": result.get("productName", ""),
            "price": result.get("price", ""),
            "asin": result.get("asin", ""),
            "rating": result.get("rating", ""),
            "product_url": result.get("productUrl", ""),
        }
        for result in json_response["parsing"]["entities"]["SearchResult"]
    ]
    return products


class NimbleAmazonBot(BaseModel):
    history: list[ChatCompletionMessageParam] = []

    @openai.call(model="gpt-3.5-turbo", tools=[amazon_search])
    @prompt_template(
        """
        SYSTEM: Answer the user's question using your `amazon_search` tool.
        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _call(self, query: str): ...

    def _step(self, query: str):
        response = self._call(query)
        self.history += [response.user_message_param, response.message_param]
        if tools := response.tools:
            for tool in tools:
                self.history += response.tool_message_params([(tool, tool.call())])
            return self._step("")
        else:
            return response.content

    def run(self):
        while True:
            query = input("(User): ")
            if query in ["exit", "quit"]:
                break
            response = self._step(query)
            print(f"(Assistant): {response}")


bot = NimbleAmazonBot()
bot.run()



import json
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel, Field

from mirascope.core.base.call_response import BaseCallResponse
from mirascope.core.base.stream import BaseStream
from mirascope.core.base.structured_stream import BaseStructuredStream
from mirascope.core.base.tool import BaseTool
from mirascope.integrations.otel import _utils


class FormatBook(BaseTool):
    """Returns the title and author of a book nicely formatted."""

    title: str = Field(..., description="The title of the book.")
    author: str = Field(..., description="The author of the book in all caps.")

    def call(self) -> str:
        return f"{self.title} by {self.author}"  # pragma: no cover


class MyCallResponse(BaseCallResponse):
    @property
    def content(self) -> str:
        return "content"  # pragma: no cover

    @property
    def tools(self) -> list[BaseTool]:
        return [
            FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")
        ]  # pragma: no cover


patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
patch.multiple(BaseStream, __abstractmethods__=set()).start()


@patch("mirascope.integrations.otel._utils.get_tracer", new_callable=MagicMock)
def test_custom_context_manager(mock_get_tracer: MagicMock) -> None:
    """Tests the `custom_context_manager` function."""
    mock_fn = MagicMock(__name__="dummy_function")
    mock_span = MagicMock(name="MockSpan")
    mock_tracer = MagicMock()
    mock_tracer.start_as_current_span.return_value = mock_span
    mock_span.__enter__.return_value = mock_span
    mock_get_tracer.return_value = mock_tracer

    with _utils.custom_context_manager(mock_fn) as span:
        mock_get_tracer.assert_called_once_with("otel")
        assert span is mock_span
        mock_tracer.start_as_current_span.assert_called_once_with("dummy_function")


@patch("mirascope.integrations.otel._utils.get_tracer", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.set_tracer_provider", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.TracerProvider", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.SimpleSpanProcessor", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.ConsoleSpanExporter", new_callable=MagicMock)
def test_configure_no_processor(
    mock_console_span_exporter: MagicMock,
    mock_simple_span_processor: MagicMock,
    mock_tracer_provider: MagicMock,
    mock_set_tracer_provider: MagicMock,
    mock_get_tracer: MagicMock,
) -> None:
    """Tests the `configure` function with no processors."""
    mock_add_span_processor = MagicMock()
    mock_tracer_provider.return_value.add_span_processor = mock_add_span_processor
    _utils.configure(None)
    mock_tracer_provider.assert_called_once()
    mock_console_span_exporter.assert_called_once()
    mock_simple_span_processor.assert_called_once_with(
        mock_console_span_exporter.return_value
    )
    mock_add_span_processor.assert_called_once_with(
        mock_simple_span_processor.return_value
    )
    mock_set_tracer_provider.assert_called_once_with(mock_tracer_provider.return_value)
    mock_get_tracer.assert_called_once_with("otel")


@patch("mirascope.integrations.otel._utils.get_tracer", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.set_tracer_provider", new_callable=MagicMock)
@patch("mirascope.integrations.otel._utils.TracerProvider", new_callable=MagicMock)
def test_configure_with_processors(
    mock_tracer_provider: MagicMock,
    mock_set_tracer_provider: MagicMock,
    mock_get_tracer: MagicMock,
) -> None:
    """Tests the `configure` function with processors."""
    processors: list = [MagicMock()]
    mock_add_span_processor = MagicMock()
    mock_tracer_provider.return_value.add_span_processor = mock_add_span_processor
    _utils.configure(processors)
    mock_tracer_provider.assert_called_once()
    mock_add_span_processor.assert_called_once_with(processors[0])
    mock_set_tracer_provider.assert_called_once_with(mock_tracer_provider.return_value)
    mock_get_tracer.assert_called_once_with("otel")


def test_get_call_response_attributes() -> None:
    """Tests the `get_call_response_attributes` function."""
    call_response = MyCallResponse(
        metadata={"tags": {"version:0001"}},
        response="hello world",
        tool_types=[],
        prompt_template="Recommend a {genre} book for me to read.",
        fn_args={"genre": "nonfiction"},
        dynamic_config={"computed_fields": {"genre": "nonfiction"}},
        messages=[],
        call_params={"tool_choice": "required"},
        call_kwargs={
            "tool_choice": "required",
            "tools": [],
            "model": "gpt-4o",
            "messages": [],
        },
        user_message_param={
            "role": "user",
            "content": [
                {"type": "text", "text": "Recommend a nonfiction book for me to read."}
            ],
        },
        start_time=100,
        end_time=200,
    )  # type: ignore
    result = _utils.get_call_response_attributes(call_response)
    assert result["gen_ai.system"] == call_response.prompt_template
    assert result["gen_ai.request.model"] == call_response.call_kwargs.get("model")
    assert result["gen_ai.request.max_tokens"] == 0
    assert result["gen_ai.request.temperature"] == 0
    assert result["gen_ai.request.top_p"] == 0
    assert result["gen_ai.response.model"] == (
        call_response.model if call_response.model else ""
    )
    assert result["gen_ai.response.id"] == (
        call_response.id if call_response.id else ""
    )
    assert result["gen_ai.response.finish_reasons"] == (
        call_response.finish_reasons if call_response.finish_reasons else ""
    )
    assert result["gen_ai.usage.completion_tokens"] == (
        call_response.output_tokens if call_response.output_tokens else ""
    )
    assert result["gen_ai.usage.prompt_tokens"] == (
        call_response.input_tokens if call_response.input_tokens else ""
    )


def test_set_call_response_event_attributes() -> None:
    """Tests the `set_call_response_event_attributes` function."""
    result = MagicMock()
    result.user_message_param = {"role": "user", "content": "user_content"}
    result.message_param = {"role": "assistant", "content": "assistant_content"}
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event

    _utils.set_call_response_event_attributes(result, span)
    assert add_event.call_count == 2
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.prompt"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.prompt"] == json.dumps(
        result.user_message_param
    )
    assert add_event.call_args_list[1][0][0] == "gen_ai.content.completion"
    assert add_event.call_args_list[1][1]["attributes"][
        "gen_ai.completion"
    ] == json.dumps(result.message_param)


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.otel._utils.set_call_response_event_attributes",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_call_response(
    mock_set_call_response_event_attributes: MagicMock,
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_call_response` function."""
    mock_fn = MagicMock()
    assert _utils.handle_call_response(MagicMock(), mock_fn, None) is None

    result = MagicMock()
    result.tools = [
        FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")
    ]
    span = MagicMock()
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    _utils.handle_call_response(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(result)
    assert mock_get_call_response_attributes.return_value["async"] is False
    mock_set_call_response_event_attributes.assert_called_once_with(result, span)


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.otel._utils.set_call_response_event_attributes",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_call_response_async(
    mock_set_call_response_event_attributes: MagicMock,
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_call_response_async` function."""
    mock_fn = MagicMock()
    assert await _utils.handle_call_response_async(MagicMock(), mock_fn, None) is None

    result = MagicMock()
    result.tools = [
        FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")
    ]
    span = MagicMock()
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    await _utils.handle_call_response_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(result)
    assert mock_get_call_response_attributes.return_value["async"] is True
    mock_set_call_response_event_attributes.assert_called_once_with(result, span)


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.otel._utils.set_call_response_event_attributes",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_stream(
    mock_set_call_response_event_attributes: MagicMock,
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_stream` function."""
    mock_fn = MagicMock()
    assert _utils.handle_stream(MagicMock(), mock_fn, None) is None

    span = MagicMock()
    set_attributes = MagicMock()
    result = MagicMock(spec=BaseStream)
    mock_construct_call_response = MagicMock()
    result.construct_call_response = mock_construct_call_response
    span.set_attributes = set_attributes
    _utils.handle_stream(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(
        mock_construct_call_response()
    )
    assert mock_get_call_response_attributes.return_value["async"] is False
    mock_set_call_response_event_attributes.assert_called_once_with(
        mock_construct_call_response(), span
    )


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.otel._utils.set_call_response_event_attributes",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_stream_async(
    mock_set_call_response_event_attributes: MagicMock,
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_stream_async` function."""
    mock_fn = MagicMock()
    assert await _utils.handle_stream_async(MagicMock(), mock_fn, None) is None

    span = MagicMock()
    set_attributes = MagicMock()
    result = MagicMock(spec=BaseStream)
    mock_construct_call_response = MagicMock()
    result.construct_call_response = mock_construct_call_response
    span.set_attributes = set_attributes
    await _utils.handle_stream_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(
        mock_construct_call_response()
    )
    assert mock_get_call_response_attributes.return_value["async"] is True
    mock_set_call_response_event_attributes.assert_called_once_with(
        mock_construct_call_response(), span
    )


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_response_model(
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_response_model` function with `BaseModel` result."""
    mock_fn = MagicMock()
    assert _utils.handle_response_model(MagicMock(), mock_fn, None) is None

    result = MagicMock(spec=BaseModel)
    response = MagicMock()
    result._response = response
    response.user_message_param = {"role": "user", "content": "user_content"}
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    _utils.handle_response_model(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(response)
    assert mock_get_call_response_attributes.return_value["async"] is False

    assert add_event.call_count == 2
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.prompt"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.prompt"] == json.dumps(
        response.user_message_param
    )
    assert add_event.call_args_list[1][0][0] == "gen_ai.content.completion"
    assert (
        add_event.call_args_list[1][1]["attributes"]["gen_ai.completion"]
        == result.model_dump_json()
    )


def test_handle_response_model_base_type() -> None:
    """Tests the `handle_response_model` function with `BaseType` result."""
    mock_fn = MagicMock()
    result = b"foo"
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    _utils.handle_response_model(result, mock_fn, span)
    assert set_attributes.call_count == 1
    set_attributes.assert_called_once_with({"async": False})
    assert add_event.call_count == 1
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.completion"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.completion"] == str(
        result
    )


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_structured_stream(
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_structured_stream` function."""
    mock_fn = MagicMock()
    assert _utils.handle_structured_stream(MagicMock(), mock_fn, None) is None

    class Foo(BaseModel):
        bar: str

    result = MagicMock(spec=BaseStructuredStream)
    response = MagicMock()
    result.stream = response
    result.constructed_response_model = Foo(bar="baz")
    response.user_message_param = {"role": "user", "content": "user_content"}
    mock_construct_call_response = MagicMock()
    response.construct_call_response = mock_construct_call_response
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    _utils.handle_structured_stream(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(
        mock_construct_call_response()
    )
    assert mock_get_call_response_attributes.return_value["async"] is False

    assert add_event.call_count == 2
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.prompt"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.prompt"] == json.dumps(
        response.user_message_param
    )
    assert add_event.call_args_list[1][0][0] == "gen_ai.content.completion"
    assert (
        add_event.call_args_list[1][1]["attributes"]["gen_ai.completion"]
        == Foo(bar="baz").model_dump_json()
    )
    result.constructed_response_model = "test"
    _utils.handle_structured_stream(result, mock_fn, span)
    assert add_event.call_args_list[3][1]["attributes"]["gen_ai.completion"] == "test"


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_response_model_async(
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_response_model_async` function with `BaseModel` result."""
    mock_fn = MagicMock()
    assert await _utils.handle_response_model_async(MagicMock(), mock_fn, None) is None

    result = MagicMock(spec=BaseModel)
    response = MagicMock()
    result._response = response
    response.user_message_param = {"role": "user", "content": "user_content"}
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    await _utils.handle_response_model_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(response)
    assert mock_get_call_response_attributes.return_value["async"] is True

    assert add_event.call_count == 2
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.prompt"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.prompt"] == json.dumps(
        response.user_message_param
    )
    assert add_event.call_args_list[1][0][0] == "gen_ai.content.completion"
    assert (
        add_event.call_args_list[1][1]["attributes"]["gen_ai.completion"]
        == result.model_dump_json()
    )


@pytest.mark.asyncio
async def test_handle_response_model_async_base_type() -> None:
    """Tests the `handle_response_model_async` function with `BaseType` result."""
    mock_fn = MagicMock()
    result = b"foo"
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    await _utils.handle_response_model_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    set_attributes.assert_called_once_with({"async": True})
    assert add_event.call_count == 1
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.completion"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.completion"] == str(
        result
    )


@patch(
    "mirascope.integrations.otel._utils.get_call_response_attributes",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_structured_stream_async(
    mock_get_call_response_attributes: MagicMock,
) -> None:
    """Tests the `handle_structured_stream_async` function."""
    mock_fn = MagicMock()
    assert (
        await _utils.handle_structured_stream_async(MagicMock(), mock_fn, None) is None
    )

    class Foo(BaseModel):
        bar: str

    result = MagicMock(spec=BaseStructuredStream)
    response = MagicMock()
    result.stream = response
    result.constructed_response_model = Foo(bar="baz")
    response.user_message_param = {"role": "user", "content": "user_content"}
    mock_construct_call_response = MagicMock()
    response.construct_call_response = mock_construct_call_response
    span = MagicMock()
    add_event = MagicMock()
    span.add_event = add_event
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    await _utils.handle_structured_stream_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_call_response_attributes.assert_called_once_with(
        mock_construct_call_response()
    )
    assert mock_get_call_response_attributes.return_value["async"] is True

    assert add_event.call_count == 2
    assert add_event.call_args_list[0][0][0] == "gen_ai.content.prompt"
    assert add_event.call_args_list[0][1]["attributes"]["gen_ai.prompt"] == json.dumps(
        response.user_message_param
    )
    assert add_event.call_args_list[1][0][0] == "gen_ai.content.completion"
    assert (
        add_event.call_args_list[1][1]["attributes"]["gen_ai.completion"]
        == Foo(bar="baz").model_dump_json()
    )
    result.constructed_response_model = "test"
    await _utils.handle_structured_stream_async(result, mock_fn, span)
    assert add_event.call_args_list[3][1]["attributes"]["gen_ai.completion"] == "test"



from tenacity import retry, stop_after_attempt, wait_exponential

from mirascope.core import anthropic, prompt_template


def format_book(title: str, author: str) -> str:
    return f"{title} by {author}"


@anthropic.call("claude-3-5-sonnet-20240620", tools=[format_book])
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def run():
    response = recommend_book("fantasy")
    if tool := response.tool:
        print(tool.call())
    else:
        print(response.content)


run()



"""Tests the `_utils.parse_content_template` function."""

from unittest.mock import MagicMock, patch

import pytest

from mirascope.core.base._utils._parse_content_template import parse_content_template
from mirascope.core.base.message_param import (
    AudioPart,
    BaseMessageParam,
    ImagePart,
    TextPart,
)


def test_parse_content_template() -> None:
    """Test the parse_content_template function."""
    assert parse_content_template("user", "", {}) is None
    assert parse_content_template("user", " ", {}) is None
    assert parse_content_template("system", "", {}) is None
    template = "This is a {var1} template with {var2} variables."
    values = {"var1": "test", "var2": "two"}
    expected = BaseMessageParam(
        role="user", content="This is a test template with two variables."
    )
    assert parse_content_template("user", template, values) == expected


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_images(
    mock_urlopen: MagicMock, mock_open: MagicMock
) -> None:
    """Test the parse_content_template function with image templates."""
    image_data = b"\xff\xd8\xffimage data"
    mock_response = MagicMock()
    mock_response.read = lambda: image_data
    mock_urlopen.return_value.__enter__.return_value = mock_response
    mock_open.return_value.__enter__.return_value = mock_response
    template = "Analyze this image: {url:image}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this image:"),
            ImagePart(
                type="image", media_type="image/jpeg", image=image_data, detail=None
            ),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected
    assert parse_content_template("user", template, {"url": "./image.jpg"}) == expected
    assert parse_content_template("user", template, {"url": image_data}) == expected

    template = "Analyze this image: {url:image(detail=low)}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this image:"),
            ImagePart(
                type="image", media_type="image/jpeg", image=image_data, detail="low"
            ),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected

    template = "Analyze these images: {urls:images}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these images:"),
            ImagePart(
                type="image", media_type="image/jpeg", image=image_data, detail=None
            ),
            ImagePart(
                type="image", media_type="image/jpeg", image=image_data, detail=None
            ),
        ],
    )
    assert (
        parse_content_template("user", template, {"urls": ["https://", "https://."]})
        == expected
    )
    assert (
        parse_content_template(
            "user", template, {"urls": ["./image.jpg", "./image.jpg"]}
        )
        == expected
    )
    assert (
        parse_content_template("user", template, {"urls": [image_data, image_data]})
        == expected
    )

    with pytest.raises(
        ValueError,
        match="When using 'images' template, 'urls' must be a list.",
    ):
        parse_content_template("user", template, {"urls": None})


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_audio(
    mock_urlopen: MagicMock, mock_open: MagicMock
) -> None:
    """Test the parse_content_template function with image templates."""
    audio_data = b"ID3audio data"
    mock_response = MagicMock()
    mock_response.read = lambda: audio_data
    mock_urlopen.return_value.__enter__.return_value = mock_response
    mock_open.return_value.__enter__.return_value = mock_response
    template = "Analyze this audio: {url:audio}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this audio:"),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected
    assert parse_content_template("user", template, {"url": "./audio.mp3"}) == expected
    assert parse_content_template("user", template, {"url": audio_data}) == expected

    template = "Analyze these audio files: {urls:audios}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these audio files:"),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
        ],
    )
    assert (
        parse_content_template("user", template, {"urls": ["https://", "https://."]})
        == expected
    )
    assert (
        parse_content_template(
            "user", template, {"urls": ["./audio.mp3", "./audio.mp3"]}
        )
        == expected
    )
    assert (
        parse_content_template("user", template, {"urls": [audio_data, audio_data]})
        == expected
    )

    with pytest.raises(
        ValueError,
        match="When using 'audios' template, 'urls' must be a list.",
    ):
        parse_content_template("user", template, {"urls": None})



from mirascope.core import anthropic, prompt_template
from mirascope.integrations.otel import configure, with_otel

configure()


def format_book(title: str, author: str) -> str:
    return f"{title} by {author}"


@with_otel()
@anthropic.call(model="claude-3-5-sonnet-20240620", tools=[format_book])
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



"""Mirascope implementation of LangGraph quickstart.

Here is the reference link to LangGraph quickstart:
    https://langchain-ai.github.io/langgraph/tutorials/introduction
"""

import requests
from bs4 import BeautifulSoup
from duckduckgo_search import DDGS
from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class WebSearch(openai.OpenAITool):
    """Search the web for the given text and parse the paragraphs of the results."""

    query: str = Field(..., description="The text to search for.")

    def call(self) -> str:
        """Search the web for the given text and parse the paragraphs of the results.

        Returns:
            Parsed paragraphs of each of the webpages, separated by newlines.
        """
        try:
            # Search the web for the given text
            results = DDGS(proxy=None).text(self.query, max_results=2)

            # Parse the paragraphs of each resulting webpage
            parsed_results = []
            for result in results:
                link = result["href"]
                try:
                    response = requests.get(link)
                    soup = BeautifulSoup(response.content, "html.parser")
                    parsed_results.append(
                        "\n".join([p.text for p in soup.find_all("p")])
                    )
                except Exception as e:
                    parsed_results.append(
                        f"{type(e)}: Failed to parse content from URL {link}"
                    )

            return "\n\n".join(parsed_results)

        except Exception as e:
            return f"{type(e)}: Failed to search the web for text"


class RequestAssistance(openai.OpenAITool):
    """A tool that requests assistance from a human expert."""

    query: str = Field(
        ...,
        description="The request for assistance needed to properly respond to the user",
    )

    def call(self) -> str:
        """Prompts a human to enter a response."""
        print(f"I am in need of assistance. {self.query}")
        response = input("\t(Human): ")
        return f"Human response: {response}"


class Chatbot(BaseModel):
    history: list[ChatCompletionMessageParam] = []

    @openai.call(model="gpt-4o-mini", stream=True, tools=[WebSearch, RequestAssistance])
    @prompt_template(
        """
        SYSTEM:
        You are an expert web searcher. 
        Your task is to answer the user's question using the provided tools.
        You have access to the following tools:
            - `WebSearch`: Search the web for information.
            - `RequestAssistance`: Request assistance from a human expert if you do not
                know how to answer the question.

        Once you have gathered all of the information you need, generate a writeup that
        strikes the right balance between brevity and completeness. The goal is to
        provide as much information to the writer as possible without overwhelming them.

        MESSAGES: {self.history}
        USER: {question}
        """
    )
    def _call(self, question: str | None = None): ...

    def _interrupt_before(self, tool: openai.OpenAITool) -> openai.OpenAITool | None:
        """Interrupt before the tool is called. Return the modified tool or None."""
        if not isinstance(tool, WebSearch):
            return tool
        response = input(f"Do you want to use the {tool._name()} tool? (y/n): ")
        if response.lower() in ["n", "no"]:
            response = input(
                f"Do you want to modify the {tool._name()} tool's query? (y/n): "
            )
            if response.lower() in ["n", "no"]:
                return None
            else:
                tool.query = input("(Assistant): Enter a new query: ")
                return tool
        else:
            return tool

    def _step(self, question: str | None = None):
        response = self._call(question)
        tools_and_outputs = []
        for chunk, tool in response:
            if tool:
                new_tool = self._interrupt_before(tool)
                if new_tool:
                    tools_and_outputs.append((new_tool, new_tool.call()))
            else:
                print(chunk.content, end="", flush=True)
        if response.user_message_param:
            self.history.append(response.user_message_param)
        self.history.append(response.message_param)
        if tools_and_outputs:
            self.history += response.tool_message_params(tools_and_outputs)
            return self._step()
        return response.content

    def run(self):
        while True:
            question = input("(User): ")
            if question in ["quit", "exit"]:
                print("(Assistant): Have a great day!")
                break
            print("(Assistant): ", end="", flush=True)
            self._step(question)
            print("")


Chatbot().run()



"""Tests the `anthropic._utils.get_json_output` module."""

import pytest
from anthropic.types import (
    Message,
    RawContentBlockDeltaEvent,
    TextBlock,
    TextDelta,
    ToolUseBlock,
    Usage,
)

try:
    from anthropic.types import (
        InputJsonDelta as InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )
except ImportError:
    from anthropic.types import (
        InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )

from mirascope.core.anthropic._utils._get_json_output import get_json_output
from mirascope.core.anthropic.call_params import AnthropicCallParams
from mirascope.core.anthropic.call_response import AnthropicCallResponse
from mirascope.core.anthropic.call_response_chunk import AnthropicCallResponseChunk


def test_get_json_output_call_response() -> None:
    """Tests the `get_json_output` function with a call response."""
    usage = Usage(input_tokens=1, output_tokens=1)
    text = TextBlock(
        type="text",
        text='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
    )
    completion = Message(
        id="id",
        content=[text],
        model="claude-3-5-sonnet-20240620",
        role="assistant",
        stop_reason="end_turn",
        stop_sequence=None,
        type="message",
        usage=usage,
    )
    call_response = AnthropicCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert (
        get_json_output(call_response, json_mode=True)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )
    completion.content[0] = ToolUseBlock(
        type="tool_use",
        id="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
    )

    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    completion.content[0] = ToolUseBlock(
        id="id", input=None, name="FormatBook", type="tool_use"
    )

    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk() -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    tool_call = InputJSONDelta(
        type="input_json_delta",
        partial_json='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
    )
    chunk = RawContentBlockDeltaEvent(
        index=0,
        delta=TextDelta(
            text="json_output",
            type="text_delta",
        ),
        type="content_block_delta",
    )
    call_response_chunk = AnthropicCallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"
    chunk.delta = tool_call
    assert (
        get_json_output(call_response_chunk, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )
    chunk.delta = TextDelta(
        type="text_delta",
        text="",
    )

    assert get_json_output(call_response_chunk, json_mode=False) == ""



from mirascope.core import openai


@openai.call("gpt-4o-mini")
def recommend_book(genre: str, style: str) -> openai.OpenAIDynamicConfig:
    return {
        "messages": [
            {"role": "system", "content": "You are a helpful book recommender."},
            {
                "role": "user",
                "content": f"Recommend a {genre} book in the style of {style}.",
            },
        ]
    }


response = recommend_book("science fiction", "cyberpunk")
print(response.content)



"""Tests the `_utils.format_template` module."""

from unittest.mock import MagicMock, patch

from mirascope.core.base._utils._format_template import format_template


@patch(
    "mirascope.core.base._utils._format_template.get_template_values",
    new_callable=MagicMock,
)
@patch(
    "mirascope.core.base._utils._format_template.get_template_variables",
    new_callable=MagicMock,
)
def test_format_template(
    mock_get_template_variables: MagicMock, mock_get_template_values: MagicMock
) -> None:
    """Tests the `format_template` function."""
    mock_get_template_variables.return_value = [("genre", None)]
    attrs = {"genre": "fantasy"}
    mock_get_template_values.return_value = attrs
    template = """
    Recommend a {genre} book.
    
    """
    formatted_template = format_template(template, attrs)
    assert formatted_template == "Recommend a fantasy book."
    mock_get_template_variables.assert_called_once_with(
        "Recommend a {genre} book.", True
    )
    mock_get_template_values.assert_called_once_with([("genre", None)], attrs)


def test_format_template_with_none_attrs() -> None:
    """Tests the `format_template` function with `None` template variables."""
    attrs = {"genre": None}
    template = """
    Recommend a book.

    {genre}
    """
    formatted_template = format_template(template, attrs)
    assert formatted_template == "Recommend a book."


def test_format_template_with_list_format_spec() -> None:
    """Tests the `format_template` function with list format spec."""
    attrs = {"genres": ["fantasy", "scifi"]}

    template = "Recommend a {genres} book."
    formatted_template = format_template(template, attrs)
    assert formatted_template == "Recommend a ['fantasy', 'scifi'] book."

    template = """
    Recommend books from one of the following genres:
    {genres:list}
    """
    formatted_template = format_template(template, attrs)
    assert (
        formatted_template
        == "Recommend books from one of the following genres:\nfantasy\nscifi"
    )



import inspect

from typing_extensions import TypedDict

from mirascope.core import openai, prompt_template


class FewShotExample(TypedDict):
    question: str
    answer: str


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Examples:
    {examples:lists}

    Query: {query}
    """
)
def self_ask(query: str, examples: list[FewShotExample]) -> openai.OpenAIDynamicConfig:
    return {
        "computed_fields": {
            "examples": [
                [example["question"], example["answer"]] for example in examples
            ]
        }
    }


few_shot_examples = [
    FewShotExample(
        question="When does monsoon season end in the state the area code 575 is located?",
        answer=inspect.cleandoc(
            """
            Are follow up questions needed here: Yes.
            Follow up: Which state is the area code 575 located in?
            Intermediate answer: The area code 575 is located in New Mexico.
            Follow up: When does monsoon season end in New Mexico?
            Intermediate answer: Monsoon season in New Mexico typically ends in mid-September.
            So the final answer is: mid-September.
            """
        ),
    ),
    FewShotExample(
        question="What is the current official currency in the country where Ineabelle Diaz is a citizen?",
        answer=inspect.cleandoc(
            """
            Are follow up questions needed here: Yes.
            Follow up: Which country is Ineabelle Diaz a citizen of?
            Intermediate answer: Ineabelle Diaz is from Peurto Rico, which is in the United States of America.
            Follow up: What is the current official currency in the United States of America?
            Intermediate answer: The current official currency in the United States is the United States dollar.
            So the final answer is: United States dollar.
            """
        ),
    ),
    FewShotExample(
        question="Where was the person who founded the American Institute of Public Opinion in 1935 born?",
        answer=inspect.cleandoc(
            """
            Are follow up questions needed here: Yes.
            Follow up: Who founded the American Institute of Public Opinion in 1935?
            Intermediate answer: George Gallup.
            Follow up: Where was George Gallup born?
            Intermediate answer: George Gallup was born in Jefferson, Iowa.
            So the final answer is: Jefferson.
            """
        ),
    ),
    FewShotExample(
        question="What language is used by the director of Tiffany Memorandum?",
        answer=inspect.cleandoc(
            """
            Are follow up questions needed here: Yes.
            Follow up: Who directed the movie called Tiffany Memorandum?
            Intermediate answer: Sergio Grieco.
            Follow up: What language is used by Sergio Grieco?
            Intermediate answer: Sergio Grieco speaks Italian.
            So the final answer is: Italian.
            """
        ),
    ),
    FewShotExample(
        question="What is the sports team the person played for who scored the first touchdown in Superbowl 1?",
        answer=inspect.cleandoc(
            """
            Are follow up questions needed here: Yes.
            Follow up: Which player scored the first touchdown in Superbowl 1?
            Intermediate answer: Max McGee.
            Follow up: Which sports team did Max McGee play for?
            Intermediate answer: Max McGee played for the Green Bay Packers.
            So the final answer is: Green Bay Packers.
            """
        ),
    ),
]

query = "The birth country of Jayantha Ketagoda left the British Empire when?"
response = self_ask(query=query, examples=few_shot_examples)
print(response.content)
# > Are follow up questions needed here: Yes.
#   Follow up: What is the birth country of Jayantha Ketagoda?
#   Intermediate answer: Jayantha Ketagoda is from Sri Lanka.
#   Follow up: When did Sri Lanka leave the British Empire?
#   Intermediate answer: Sri Lanka, formerly known as Ceylon, gained independence from the British Empire on February 4, 1948.
#   So the final answer is: February 4, 1948.

response = self_ask(query=query, examples=[])
print(response.content)
# > Jayantha Ketagoda was born in Sri Lanka, which was known as Ceylon during the
#   British colonial period. Ceylon gained independence from the British Empire on
#   February 4, 1948.



from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini", stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


for chunk, _ in recommend_book("fantasy"):
    content = chunk.chunk.choices[0].delta.content
    print(content, end="", flush=True)



import asyncio

from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class Skeleton(BaseModel):
    subpoints: list[str] = Field(
        ...,
        description="""The subpoints of the skeleton of the original query.
        Each is 3-5 words and starts with its point index, e.g. 
        1. Some subpoint...""",
    )


@openai.call(model="gpt-3.5-turbo", response_model=Skeleton)
@prompt_template(
    """
    You’re an organizer responsible for only giving the skeleton
    (not the full content) for answering the question.
    Provide the skeleton in a list of points (numbered 1., 2., 3., etc.) to
    answer the question. Instead of writing a full sentence, each skeleton point
    should be very short with only 3∼5 words. Generally, the skeleton should have
    3∼10 points. Now, please provide the skeleton for the following question.
    {query}
    Skeleton:
    """
)
def break_into_subpoints(query: str): ...


@openai.call(model="gpt-3.5-turbo")
@prompt_template(
    """
    You’re responsible for continuing the writing of one and only one point in
    the overall answer to the following question:

    {query}

    The skeleton of the answer is:

    {skeleton}

    Continue and only continue the writing of point {point_index}. Write it very
    shortly in 1-2 sentences and do not continue with other points!
    """
)
async def expand_subpoint(query: str, skeleton: list[str], point_index: int): ...


query = "How can I improve my focus?"


async def skeleton_of_thought(query):
    skeleton = break_into_subpoints(query)
    # Uncomment to see intermediate responses
    # print(skeleton)
    tasks = [
        expand_subpoint(query, skeleton.subpoints, i + 1)
        for i, subpoint in enumerate(skeleton.subpoints)
    ]
    results = await asyncio.gather(*tasks)
    return "\n".join([result.content for result in results])


print(asyncio.run(skeleton_of_thought(query)))
# > 1. Set specific goals: Clearly define what you want to achieve and break down your tasks into manageable steps to stay focused on your objectives.
#   2. Create a schedule: Organize your tasks and allocate specific times to work on them, allowing you to stay on track and prioritize effectively.
#   3. Minimize distractions by turning off notifications, finding a quiet workspace, and setting boundaries with others.
#   4. Taking regular breaks throughout the day can help improve focus by giving your mind a chance to rest and rejuvenate, allowing you to come back to your tasks with renewed energy and clarity.
#   5. Practice mindfulness by being fully present in the moment and focusing on one task at a time to improve your concentration and mental clarity.

# Intermediate Responses

# skeleton = break_into_subpoints(prompt)
# > subpoints=['1. Set specific goals', '2. Create a schedule', '3. Minimize distractions', '4. Take breaks', '5. Practice mindfulness']



"""The `OpenAITool` class for easy tool usage with OpenAI LLM calls.

usage docs: learn/tools.md#using-tools-with-standard-calls
"""

from __future__ import annotations

import copy

from anthropic.types import ToolParam, ToolUseBlock
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class AnthropicTool(BaseTool):
    """A class for defining tools for Anthropic LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.anthropic import anthropic_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @anthropic_call("claude-3-5-sonnet-20240620", tools=[format_book])
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `AnthropicTool` instance
        print(tool.call())
    ```
    """

    tool_call: SkipJsonSchema[ToolUseBlock]

    @classmethod
    def tool_schema(cls) -> ToolParam:
        """Constructs a `ToolParam` tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.anthropic import AnthropicTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = AnthropicTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Anthropic-specific tool schema
        ```
        """
        return ToolParam(
            input_schema=cls.model_tool_schema(),
            name=cls._name(),
            description=cls._description(),
        )

    @classmethod
    def from_tool_call(cls, tool_call: ToolUseBlock) -> AnthropicTool:
        """Constructs an `AnthropicTool` instance from a `tool_call`.

        Args:
            tool_call: The Anthropic tool call from which to construct this tool
                instance.
        """
        model_json = copy.deepcopy(tool_call.input)
        model_json["tool_call"] = tool_call.model_dump()  # type: ignore
        return cls.model_validate(model_json)



"""Beta implementation of the new OpenAI parse functionality."""

import inspect
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import Literal, ParamSpec, Protocol, TypeVar, cast, overload

from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel

from mirascope.core.base import BaseMessageParam, _utils
from mirascope.core.openai import OpenAICallParams, OpenAIDynamicConfig, OpenAITool
from mirascope.core.openai._utils import convert_message_params

_ResponseModelT = TypeVar("_ResponseModelT", bound=_utils.BaseType | BaseModel)
_P = ParamSpec("_P")


class ParseFunctionDecorator(Protocol[_ResponseModelT]):
    @overload
    def __call__(self, fn: Callable[..., OpenAIDynamicConfig]) -> _ResponseModelT: ...

    @overload
    def __call__(
        self, fn: Callable[..., Awaitable[OpenAIDynamicConfig]]
    ) -> Awaitable[_ResponseModelT]: ...

    def __call__(
        self, fn: Callable[..., OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]]
    ) -> _ResponseModelT | Awaitable[_ResponseModelT]: ...


def parse(
    model: Literal["gpt-4o-2024-08-06"],
    response_model: type[_ResponseModelT],
    call_params: OpenAICallParams | None = None,
) -> ParseFunctionDecorator[_ResponseModelT]:
    """Beta implementation of the new OpenAI parse functionality."""

    @overload
    def decorator(
        fn: Callable[_P, OpenAIDynamicConfig],
    ) -> Callable[_P, _ResponseModelT]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[OpenAIDynamicConfig]],
    ) -> Callable[_P, Awaitable[_ResponseModelT]]: ...

    def decorator(
        fn: Callable[_P, OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    ):
        if inspect.iscoroutinefunction(fn):

            @wraps(fn)
            async def inner_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> _ResponseModelT:
                assert _utils.SetupCall.fn_is_async(fn)
                fn_args = _utils.get_fn_args(fn, args, kwargs)
                dynamic_config = await fn(*args, **kwargs)
                _, messages, _, call_kwargs = _utils.setup_call(
                    fn,
                    fn_args,
                    dynamic_config,
                    None,
                    OpenAITool,
                    call_params or OpenAICallParams(),
                )
                messages = cast(
                    list[BaseMessageParam | ChatCompletionMessageParam], messages
                )
                messages = convert_message_params(messages)
                messages.append(
                    {"role": "user", "content": _utils.json_mode_content(None)}
                )
                client = AsyncOpenAI()
                call_kwargs.pop("response_format", None)
                completion = await client.beta.chat.completions.parse(
                    model=model,
                    messages=messages,
                    response_format=response_model,
                    **call_kwargs,
                )
                message = completion.choices[0].message
                if message.parsed:
                    return message.parsed
                raise RuntimeError(f"{message.refusal}")

            return inner_async
        else:

            @wraps(fn)
            def inner(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
                assert _utils.SetupCall.fn_is_sync(fn)
                fn_args = _utils.get_fn_args(fn, args, kwargs)
                dynamic_config = fn(*args, **kwargs)
                _, messages, _, call_kwargs = _utils.setup_call(
                    fn,
                    fn_args,
                    dynamic_config,
                    None,
                    OpenAITool,
                    call_params or OpenAICallParams(),
                )
                messages = cast(
                    list[BaseMessageParam | ChatCompletionMessageParam], messages
                )
                messages = convert_message_params(messages)
                messages.append(
                    {"role": "user", "content": _utils.json_mode_content(None)}
                )
                client = OpenAI()
                call_kwargs.pop("response_format", None)
                completion = client.beta.chat.completions.parse(
                    model=model,
                    messages=messages,
                    response_format=response_model,
                    **call_kwargs,
                )
                message = completion.choices[0].message
                if message.parsed:
                    return message.parsed
                raise RuntimeError(f"{message.refusal}")

            return inner

    return decorator



"""The `create_factory` method for generating provider specific create decorators."""

import datetime
import inspect
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import ParamSpec, TypeVar, overload

from ._utils import (
    SetupCall,
    get_fn_args,
    get_metadata,
    get_possible_user_message_param,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_P = ParamSpec("_P")


def create_factory(  # noqa: ANN202
    *,
    TCallResponse: type[_BaseCallResponseT],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
):
    """Returns the wrapped function with the provider specific interfaces."""

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _BaseCallResponseT | _ParsedOutputT]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]: ...

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        _BaseCallResponseT
        | _ParsedOutputT
        | Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]:
        if inspect.iscoroutinefunction(fn):

            @wraps(fn)
            async def inner_async(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> TCallResponse | _ParsedOutputT:
                assert SetupCall.fn_is_async(fn)
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = await fn(*args, **kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                    model=model,
                    client=client,
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )
                start_time = datetime.datetime.now().timestamp() * 1000
                response = await create(stream=False, **call_kwargs)
                end_time = datetime.datetime.now().timestamp() * 1000
                output = TCallResponse(
                    metadata=get_metadata(fn, dynamic_config),
                    response=response,
                    tool_types=tool_types,  # type: ignore
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                    user_message_param=get_possible_user_message_param(messages),
                    start_time=start_time,
                    end_time=end_time,
                )
                output._model = model
                return output if not output_parser else output_parser(output)

            return inner_async
        else:

            @wraps(fn)
            def inner(
                *args: _P.args, **kwargs: _P.kwargs
            ) -> TCallResponse | _ParsedOutputT:
                assert SetupCall.fn_is_sync(fn)
                fn_args = get_fn_args(fn, args, kwargs)
                dynamic_config = fn(*args, **kwargs)
                create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                    model=model,
                    client=client,
                    fn=fn,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    tools=tools,
                    json_mode=json_mode,
                    call_params=call_params,
                    extract=False,
                )
                start_time = datetime.datetime.now().timestamp() * 1000
                response = create(stream=False, **call_kwargs)
                end_time = datetime.datetime.now().timestamp() * 1000
                output = TCallResponse(
                    metadata=get_metadata(fn, dynamic_config),
                    response=response,
                    tool_types=tool_types,  # type: ignore
                    prompt_template=prompt_template,
                    fn_args=fn_args,
                    dynamic_config=dynamic_config,
                    messages=messages,
                    call_params=call_params,
                    call_kwargs=call_kwargs,
                    user_message_param=get_possible_user_message_param(messages),
                    start_time=start_time,
                    end_time=end_time,
                )
                output._model = model
                return output if not output_parser else output_parser(output)

            return inner

    return decorator



import os
from unittest.mock import MagicMock, patch

import pytest

from mirascope.integrations.otel._with_hyperdx import with_hyperdx


@pytest.fixture
def mock_env_variable():
    with patch.dict(os.environ, {"HYPERDX_API_KEY": "test-api-key"}):
        yield


@patch(
    "mirascope.integrations.otel._with_hyperdx.with_otel",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.otel._with_hyperdx.trace",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.otel._with_hyperdx.configure",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.otel._with_hyperdx.BatchSpanProcessor",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.otel._with_hyperdx.OTLPSpanExporter", new_callable=MagicMock
)
def test_with_hyperdx(
    mock_otlp_span_exporter: MagicMock,
    mock_batch_span_processor: MagicMock,
    mock_configure: MagicMock,
    mock_trace: MagicMock,
    mock_with_otel: MagicMock,
    mock_env_variable,
) -> None:
    """Tests the `with_hyperdx` decorator."""
    mock_trace.get_tracer_provider.return_value = MagicMock()
    mock_fn = MagicMock()
    with_hyperdx()(mock_fn)
    mock_with_otel.assert_called_once()
    mock_otlp_span_exporter.assert_called_once_with(
        endpoint="https://in-otel.hyperdx.io/v1/traces",
        headers={"authorization": "test-api-key"},
    )
    mock_batch_span_processor.assert_called_once_with(
        mock_otlp_span_exporter.return_value
    )
    mock_configure.assert_called_once_with(
        processors=[mock_batch_span_processor.return_value]
    )



from mirascope.core import BaseToolKit, openai, prompt_template, toolkit_tool


class BookToolkit(BaseToolKit):
    genre: str

    @toolkit_tool
    def format_book(self, title: str, author: str) -> str:
        """Format a {self.genre} book recommendation."""
        return f"{title} by {author} ({self.genre})"


@openai.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str) -> openai.OpenAIDynamicConfig:
    toolkit = BookToolkit(genre=genre)
    return {"tools": toolkit.create_tools()}


response = recommend_book("mystery")
if response.tool:
    print(response.tool.call())



from enum import Enum

from pydantic import BaseModel

from mirascope.core import openai, prompt_template


class Sentiment(Enum):
    NEGATIVE = "negative"
    NEUTRAL = "neutral"
    POSITIVE = "positive"


class SentimentWithReasoning(BaseModel):
    reasoning: str
    sentiment: Sentiment


@openai.call("gpt-4o-mini", response_model=SentimentWithReasoning)
@prompt_template(
    """
    Classify the sentiment of the following text: {text}.
    Explain your reasoning for the classified sentiment.
    """
)
def classify_sentiment_with_reasoning(text: str): ...


text = "I would recommend this product if it were cheaper..."
response = classify_sentiment_with_reasoning(text)
print(response.sentiment)
# > Sentiment.NEUTRAL
print(f"Reasoning: {response.reasoning}")
# > Reasoning: The statement expresses a conditional recommendation that is
#   dependent on the price of the product. While the speaker has a positive
#   inclination to recommend the product, the condition of it being cheaper
#   introduces a level of uncertainty, leading to a neutral sentiment overall.



"""Tests the `anthropic._utils.convert_message_params` function."""

import pytest
from anthropic.types import MessageParam

from mirascope.core.anthropic._utils._convert_message_params import (
    convert_message_params,
)
from mirascope.core.base import AudioPart, BaseMessageParam, ImagePart, TextPart


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[BaseMessageParam | MessageParam] = [
        {"role": "user", "content": [{"type": "text", "text": "Hello"}]},
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail="auto"
                ),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        {"role": "user", "content": [{"type": "text", "text": "Hello"}]},
        {"role": "user", "content": "Hello"},
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "Hello"},
                {
                    "type": "image",
                    "source": {
                        "data": "aW1hZ2U=",
                        "media_type": "image/jpeg",
                        "type": "base64",
                    },
                },
            ],
        },
    ]

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Anthropic currently only "
        "supports JPEG, PNG, GIF, and WebP images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="Anthropic currently only supports text and image modalities. "
        "Modality provided: audio",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(type="audio", media_type="audio/mp3", audio=b"audio")
                    ],
                )
            ]
        )



from mirascope.core import openai, prompt_template


@openai.call(model="gpt-3.5-turbo")
@prompt_template("Summarize this text: {text}")
def summarize(text: str): ...


@openai.call(model="gpt-3.5-turbo")
@prompt_template("Translate this text to {language}: {summary}")
def translate(summary: str, language: str): ...


def summarize_and_translate(original_text: str):
    summary = summarize(original_text)
    translation = translate(summary.content, "french")
    return translation.content



"""Tests the `cohere.tool` module."""

from cohere.types import Tool, ToolCall, ToolParameterDefinitionsValue
from pydantic import Field

from mirascope.core.base.tool import BaseTool
from mirascope.core.cohere.tool import CohereTool


def test_cohere_tool() -> None:
    """Tests the `CohereTool` class."""

    class FormatBook(CohereTool):
        """Returns the title and author nicely formatted."""

        title: str = Field(..., description="Title")
        author: str = Field(..., description="Author")

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ToolCall(
        name="FormatBook",
        parameters={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, CohereTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"
    title = ToolParameterDefinitionsValue(
        description="Title", type="string", required=True
    )
    author = ToolParameterDefinitionsValue(
        description="Author", type="string", required=True
    )
    assert FormatBook.tool_schema() == Tool(
        name="FormatBook",
        description="Returns the title and author nicely formatted.",
        parameter_definitions={"title": title, "author": author},
    )



from pydantic import BaseModel

from mirascope.core import anthropic, prompt_template


class Book(BaseModel):
    title: str
    author: str


def parse_book_recommendation(response: anthropic.AnthropicCallResponse) -> Book:
    title, author = response.content.split(" by ")
    return Book(title=title, author=author)


@anthropic.call(
    model="claude-3-5-sonnet-20240620", output_parser=parse_book_recommendation
)
@prompt_template("Recommend a {genre} book in the format Title by Author")
def recommend_book(genre: str): ...


book = recommend_book("science fiction")
print(f"Title: {book.title}")
print(f"Author: {book.author}")



from mirascope.core import openai, prompt_template


@openai.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str, creativity: float) -> openai.OpenAIDynamicConfig:
    return {"call_params": {"temperature": creativity}}


response = recommend_book("mystery", 0.8)
print(response.content)



from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Name a scientist who is well-known for their work in {field_of_study}.
    Give me just the name.
    """
)
def scientist_selector(field_of_study: str): ...


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Imagine that you are scientist {scientist}.
    Your task is to explain a theory that you, {scientist}, are famous for.
    """
)
def theory_explainer(scientist: str, theory: str): ...


scientist = scientist_selector(field_of_study="physics").content
print(scientist)
# > Albert Einstein.

theory_explanation = theory_explainer(scientist=scientist, theory="relativity").content
print(theory_explanation)
# > Certainly! Here's an explanation of the theory of relativity: ...



from enum import Enum

from mirascope.core import openai, prompt_template


class Sentiment(str, Enum):
    POSITIVE = "positive"
    NEGATIVE = "negative"


@openai.call(model="gpt-4o", response_model=Sentiment)
@prompt_template("Is the following review positive or negative? {review}")
def sentiment_classifier(review: str): ...


@openai.call("gpt-4o-mini")
@prompt_template(
    """
    SYSTEM:
    Your task is to respond to a review.
    The review has been identified as {sentiment}.
    Please write a {conditional_review_prompt}.

    USER: Write a response for the following review: {review}
    """
)
def review_responder(review: str) -> openai.OpenAIDynamicConfig:
    sentiment = sentiment_classifier(review=review)
    conditional_review_prompt = (
        "thank you response for the review."
        if sentiment == Sentiment.POSITIVE
        else "response addressing the review."
    )
    return {
        "computed_fields": {
            "conditional_review_prompt": conditional_review_prompt,
            "sentiment": sentiment,
        }
    }


positive_review = "This tool is awesome because it's so flexible!"
response = review_responder(review=positive_review)
print(response)
print(response.dynamic_config)



from typing import Annotated

from dotenv import load_dotenv
from pydantic import AfterValidator, BaseModel, Field, ValidationError
from tenacity import retry, stop_after_attempt

from mirascope.core import anthropic, prompt_template
from mirascope.integrations.tenacity import collect_errors

load_dotenv()


class SpellingAndGrammarCheck(BaseModel):
    has_errors: bool = Field(
        description="Whether the text has typos or grammatical errors"
    )


@anthropic.call(
    model="claude-3-5-sonnet-20240620",
    response_model=SpellingAndGrammarCheck,
    json_mode=True,
)
@prompt_template(
    """
    Does the following text have any typos or grammatical errors? {text}
    """
)
def check_for_errors(text: str): ...


class GrammarCheck(BaseModel):
    text: Annotated[
        str,
        AfterValidator(
            lambda t: t
            if not (check_for_errors(t)).has_errors
            else (_ for _ in ()).throw(ValueError("Text still contains errors"))
        ),
    ] = Field(description="The corrected text with proper grammar")
    explanation: str = Field(description="Explanation of the corrections made")


@retry(stop=stop_after_attempt(3), after=collect_errors(ValidationError))
@anthropic.call(
    "claude-3-5-sonnet-20240620", response_model=GrammarCheck, json_mode=True
)
@prompt_template(
    """
    {previous_errors}

    Correct the grammar in the following text.
    If no corrections are needed, return the original text.
    Provide an explanation of any corrections made.

    Text: {text}
    """
)
def correct_grammar(
    text: str, *, errors: list[ValidationError] | None = None
) -> anthropic.AnthropicDynamicConfig:
    previous_errors = f"Previous Errors: {errors}" if errors else "No previous errors."
    return {"computed_fields": {"previous_errors": previous_errors}}


try:
    text = "I has went to the store yesterday and buyed some milk."
    result = correct_grammar(text)
    print(f"Corrected text: {result.text}")
    print(f"Explanation: {result.explanation}")
except ValidationError:
    print("Failed to correct grammar after 3 attempts")



"""Tests the `anthropic.tool` module."""

from anthropic.types import (
    ToolParam,
    ToolUseBlock,
)

from mirascope.core.anthropic.tool import AnthropicTool
from mirascope.core.base.tool import BaseTool


def test_anthropic_tool() -> None:
    """Tests the `AnthropicTool` class."""

    class FormatBook(AnthropicTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ToolUseBlock(
        id="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
        type="tool_use",
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, AnthropicTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"
    assert FormatBook.tool_schema() == ToolParam(
        name="FormatBook",
        description="Returns the title and author nicely formatted.",
        input_schema={
            "properties": {
                "title": {"type": "string"},
                "author": {"type": "string"},
            },
            "required": ["title", "author"],
            "type": "object",
        },
    )



from mirascope.core import openai, prompt_template

url = "https://c02.purpledshub.com/uploads/sites/41/2023/01/How-to-see-the-Wolf-Moon-in-2023--4bb6bb7.jpg?w=940&webp=1"


@openai.call(model="gpt-4o-mini")
@prompt_template("Generate a short, descriptive caption for this image: {url:image}")
def generate_caption(url: str): ...


response = generate_caption(url)
print(response)



from mirascope.core import openai, prompt_template


@openai.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book with a reading level of {reading_level}")
def recommend_book(genre: str, age: int) -> openai.OpenAIDynamicConfig:
    reading_level = "adult"
    if age < 12:
        reading_level = "elementary"
    elif age < 18:
        reading_level = "young adult"
    return {"computed_fields": {"reading_level": reading_level}}


response = recommend_book("fantasy", 15)
print(response.content)



from pydantic import BaseModel

from mirascope.core import openai, prompt_template


class Book(BaseModel):
    title: str
    author: str


class Librarian(BaseModel):
    books: list[Book]

    @openai.call("gpt-4o-mini")
    @prompt_template(
        """
        SYSTEM: You can recommend the following books: {self.books}
        USER: I just read {previous_book.title} by {previous_book.author}. What should I read next?
        """
    )
    def recommend_book(self, previous_book: Book): ...



from mirascope.core import openai
from mirascope.core.base.prompt import prompt_template


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    The following is a sequence of events:
    {story}
    What events does {name} know about?
    """
)
def get_one_perspective(story: str, name: str):
    """Gets one person's perspective of a story."""


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    {story_from_perspective}
    Based on the above information, answer the following question:
    {query}
    """
)
def sim_to_m(story: str, query: str, name: str) -> openai.OpenAIDynamicConfig:
    """Executes the flow of the Sim to M technique."""
    story_from_perspective = get_one_perspective(story=story, name=name)
    # Uncomment to see intermediate responses
    # print(story_from_perspective)
    return {"computed_fields": {"story_from_perspective": story_from_perspective}}


story = """Jim put the ball in the box. While Jim wasn't looking, Avi moved the
ball to the basket."""
query = "Where does Jim think the ball is?"

print(sim_to_m(story=story, query=query, name="Jim"))
# > Based on the information provided, Jim thinks the ball is in the box, as he is only aware of his action of putting the ball there and does not know that Avi moved it to the basket.


# Intermediate Responses

# story_from_perspective = get_one_perspective(story=story, name=name)
# > Jim knows about the event of putting the ball in the box. However, he is unaware of Avi moving the ball to the basket because he wasn't looking at that moment. Therefore, Jim only knows about the first event: him putting the ball in the box.



from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template

few_shot_examples = """
Q: The median age in the city was 22.1 years. 10.1% of residents were under the age of 18; 56.2%
were between the ages of 18 and 24; 16.1% were from 25 to 44; 10.5% were from 45 to 64; and 7%
were 65 years of age or older. Which age group is larger: under the age of 18 or 18 and 24?
A: To answer the question “Which age group is larger: under the age of 18 or 18 and 24?”, we need
to know: “How many percent were under the age of 18?”, “How many percent were between the
ages of 18 and 24?”.
Q: Old age pensions were raised by 300 francs per month to 1,700 francs for a single person and
to 3,700 francs for a couple, while health insurance benefits were made more widely available to
unemployed persons and part-time employees. How many francs were the old age pensions for a
single person before they were raised?
A: To answer the question “How many francs were the old age pensions for a single person before
they were raised?”, we need to know: “How many francs were the old age pensions for a single
person?”, “How many francs were old age pensions raised for a single person?”.
Q: In April 2011, the ECB raised interest rates for the first time since 2008 from 1% to 1.25%, with
a further increase to 1.50% in July 2011. However, in 2012-2013 the ECB lowered interest rates to
encourage economic growth, reaching the historically low 0.25% in November 2013. Soon after the
rates were cut to 0.15%, then on 4 September 2014 the central bank reduced the rates from 0.15%
to 0.05%, the lowest rates on record. How many percentage points did interest rates drop between
April 2011 and September 2014?
A: To answer the question “How many percentage points did interest rates drop between April 2011
and September 2014?”, we need to know: “What was the interest rate in April 2011?”, “What was
the interest rate in September 2014?”.
Q: Non-nationals make up more than half of the population of Bahrain. According to government
statistics dated between 2005-2009 roughly 290,000 Indians, 125,000 Bangladeshis, 45,000 Pakistanis, 45,000 Filipinos, and 8,000 Indonesians. How many Pakistanis and Indonesians are in Bahrain?
A: To answer the question “How many Pakistanis and Indonesians are in Bahrain?”, we need to
know: “How many Pakistanis are in Bahrain?”, “How many Indonesians are in Bahrain?”.
"""


class Problem(BaseModel):
    subproblems: list[str] = Field(
        ..., description="The subproblems that the original problem breaks down into"
    )


@openai.call(model="gpt-4o-mini", response_model=Problem)
@prompt_template(
    """
    Examples to guide your answer:
    {examples}
    Break the following query into subproblems:
    {query}
    """
)
def break_into_subproblems(query: str) -> openai.OpenAIDynamicConfig:
    return {"computed_fields": {"examples": few_shot_examples}}


@openai.call(model="gpt-4o-mini")
def call(history: list[ChatCompletionMessageParam]):
    """
    MESSAGES: {history}
    """


query_context = """The Census Bureaus 2006-2010 American Community Survey showed that
(in 2010 inflation adjustment dollars) median household income was $52,056 and the
median family income was $58,942."""

query_question = "How many years did the Census Bureaus American Community Survey last?"


def least_to_most(query_context: str, query_question: str):
    problem = break_into_subproblems(query_context + query_question)
    # Uncomment to see intermediate steps
    # print(problem.subproblems)
    history: list[ChatCompletionMessageParam] = [
        {"role": "user", "content": query_context + problem.subproblems[0]}
    ]
    response = call(history=history)
    # print(response)
    history.append(response.message_param)
    if len(problem.subproblems) == 1:
        return response
    else:
        for i in range(1, len(problem.subproblems)):
            history.append({"role": "user", "content": problem.subproblems[i]})
            response = call(history=history)
            # print(response)
            history.append(response.message_param)
        return response


print(least_to_most(query_context=query_context, query_question=query_question))
# > There are 5 years from the starting year (2006) to the ending year (2010), inclusive. The years are 2006, 2007, 2008, 2009, and 2010.


# Intermediate Responses

# problem = break_into_subproblems(query_context + query_question)
# problem.subproblems
# > ["What were the starting and ending years of the Census Bureau's American Community Survey?", 'How many years are there from the starting year to the ending year?']

# response = call(history=history)
# > The American Community Survey (ACS) conducted by the Census Bureau began in 2005. Therefore, for the 2006-2010 ACS, the starting year is 2006, and the ending year is 2010.
# > There are 5 years from the starting year (2006) to the ending year (2010), inclusive. The years are 2006, 2007, 2008, 2009, and 2010.



import matplotlib.pyplot as plt
import networkx as nx
from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class Edge(BaseModel):
    source: str = Field(..., description="The source node of the edge")
    target: str = Field(..., description="The target node of the edge")
    relationship: str = Field(
        ..., description="The relationship between the source and target nodes"
    )


class Node(BaseModel):
    id: str = Field(..., description="The unique identifier of the node")
    type: str = Field(..., description="The type or label of the node")
    properties: dict | None = Field(
        ..., description="Additional properties and metadata associated with the node"
    )


class KnowledgeGraph(BaseModel):
    nodes: list[Node] = Field(..., description="List of nodes in the knowledge graph")
    edges: list[Edge] = Field(..., description="List of edges in the knowledge graph")


question = "What are the pitfalls of using LLMs?"


@openai.call(model="gpt-4o-mini", response_model=KnowledgeGraph)
@prompt_template(
    """
    SYSTEM:
    Your job is to create a knowledge graph based on the text and user question.
    
    The article:
    {text}

    Example:
    John and Jane Doe are siblings. Jane is 25 and 5 years younger than John.
    Node(id="John Doe", type="Person", properties={{"age": 30}})
    Node(id="Jane Doe", type="Person", properties={{"age": 25}})
    Edge(source="John Doe", target="Jane Doe", relationship="Siblings")

    USER:
    {question}
    """
)
def generate_knowledge_graph(
    question: str, file_name: str
) -> openai.OpenAIDynamicConfig:
    text = ""
    with open(file_name) as f:
        text = f.read()
    return {"computed_fields": {"text": text}}


kg = generate_knowledge_graph(question, "../wikipedia.txt")


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    SYSTEM:
    Answer the following question based on the knowledge graph.

    Knowledge Graph:
    {knowledge_graph}
    
    USER:
    {question}
    """
)
def run(question: str, knowledge_graph: KnowledgeGraph): ...


print(kg)
print(run(question, kg))


def render_graph(kg: KnowledgeGraph):
    G = nx.DiGraph()

    for node in kg.nodes:
        G.add_node(node.id, label=node.type, **(node.properties or {}))

    for edge in kg.edges:
        G.add_edge(edge.source, edge.target, label=edge.relationship)

    plt.figure(figsize=(15, 10))
    pos = nx.spring_layout(G)

    nx.draw_networkx_nodes(G, pos, node_size=2000, node_color="lightblue")
    nx.draw_networkx_edges(G, pos, arrowstyle="->", arrowsize=20)
    nx.draw_networkx_labels(G, pos, font_size=12, font_weight="bold")

    edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_color="red")

    plt.title("Knowledge Graph Visualization", fontsize=15)
    plt.show()


question = "What are the pitfalls of using LLMs?"
render_graph(kg)



import asyncio
import os

from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel, ConfigDict
from slack_bolt.adapter.socket_mode.async_handler import AsyncSocketModeHandler
from slack_bolt.async_app import AsyncApp

from mirascope.core import openai, prompt_template

os.environ["OPENAI_API_KEY"] = "sk-YOUR_OPENAI_API_KEY"
SLACK_BOT_TOKEN = "xoxb-..."
SLACK_APP_TOKEN = "xapp-..."


class MiraBot(BaseModel):
    app: AsyncApp = AsyncApp(token=SLACK_BOT_TOKEN)
    history: list[ChatCompletionMessageParam] = []

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    @openai.call(model="gpt-4o", call_params={"temperature": 0})
    @prompt_template(
        """
        SYSTEM:
        You are an assistant that can help with a wide range of tasks and provide
        valuable insights and information on a wide range of topics.

        {self.history}

        USER:
        {input}
        """
    )
    async def _step(self, input: str): ...

    async def register_event_handlers(self):
        @self.app.event("message")
        async def handle_message_events(message, say, logger):
            response = await self._step(message["text"])
            if response.user_message_param:
                self.history.append(response.user_message_param)
            self.history.append(response.message_param)
            await say(response.content)

    async def start(self):
        handler = AsyncSocketModeHandler(self.app, SLACK_APP_TOKEN)
        await self.register_event_handlers()
        await handler.start_async()


if __name__ == "__main__":
    asyncio.run(MiraBot().start())



"""Tests the internal `_create` module."""

from functools import partial
from typing import TypeVar, cast
from unittest.mock import MagicMock, patch

import pytest

from mirascope.core.base._create import create_factory
from mirascope.core.base.call_response import BaseCallResponse

_T = TypeVar("_T")


@pytest.fixture()
def mock_create_decorator_kwargs() -> dict:
    """Returns the mock kwargs (excluding fn) for the create `decorator` function."""

    def output_parser(output: _T) -> _T:
        return output

    return {
        "model": "model",
        "tools": [],
        "output_parser": output_parser,
        "json_mode": True,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }


@patch(
    "mirascope.core.base._create.get_possible_user_message_param",
    new_callable=MagicMock,
)
@patch("mirascope.core.base._create.get_metadata", new_callable=MagicMock)
def test_create_factory_sync(
    mock_get_metadata: MagicMock,
    mock_get_possible_user_message_param: MagicMock,
    mock_setup_call: MagicMock,
    mock_create_decorator_kwargs: dict,
) -> None:
    """Tests the `create_factory` method on a sync function."""
    (
        mock_create,
        mock_prompt_template,
        mock_messages,
        mock_tool_types,
        mock_call_kwargs,
    ) = mock_setup_call.return_value
    mock_create = cast(MagicMock, mock_create)

    decorator = partial(
        create_factory(TCallResponse=MagicMock, setup_call=mock_setup_call),
        **mock_create_decorator_kwargs,
    )

    dynamic_config = MagicMock()

    def fn(genre: str, *, topic: str):
        """Recommend a {genre} book on {topic}."""
        return dynamic_config

    output: BaseCallResponse = decorator(fn)("fantasy", topic="magic")  # type: ignore

    assert output.metadata == mock_get_metadata.return_value
    assert output.response == mock_create.return_value
    assert output.tool_types == mock_tool_types
    assert output.prompt_template == mock_prompt_template
    fn_args = {"genre": "fantasy", "topic": "magic"}
    assert output.fn_args == fn_args
    assert output.messages == mock_messages
    assert output.call_params == mock_create_decorator_kwargs["call_params"]
    assert output.call_kwargs == mock_call_kwargs
    assert (
        output.user_message_param == mock_get_possible_user_message_param.return_value
    )
    assert output.start_time is not None
    assert output.end_time is not None
    assert output._model == "model"

    mock_setup_call.assert_called_once_with(
        model="model",
        client=mock_create_decorator_kwargs["client"],
        fn=fn,
        fn_args=fn_args,
        dynamic_config=dynamic_config,
        tools=mock_create_decorator_kwargs["tools"],
        json_mode=mock_create_decorator_kwargs["json_mode"],
        call_params=mock_create_decorator_kwargs["call_params"],
        extract=False,
    )
    mock_create.assert_called_once_with(stream=False, **mock_call_kwargs)


@patch(
    "mirascope.core.base._create.get_possible_user_message_param",
    new_callable=MagicMock,
)
@patch("mirascope.core.base._create.get_metadata", new_callable=MagicMock)
@pytest.mark.asyncio
async def test_create_factory_async(
    mock_get_metadata: MagicMock,
    mock_get_possible_user_message_param: MagicMock,
    mock_setup_call_async: MagicMock,
    mock_create_decorator_kwargs: dict,
) -> None:
    """Tests the `create_factory` method on a sync function."""
    (
        mock_create,
        mock_prompt_template,
        mock_messages,
        mock_tool_types,
        mock_call_kwargs,
    ) = mock_setup_call_async.return_value
    mock_create = cast(MagicMock, mock_create)

    decorator = partial(
        create_factory(TCallResponse=MagicMock, setup_call=mock_setup_call_async),
        **mock_create_decorator_kwargs,
    )

    dynamic_config = MagicMock()

    async def fn(genre: str, *, topic: str):
        """Recommend a {genre} book on {topic}."""
        return dynamic_config

    output: BaseCallResponse = await decorator(fn)("fantasy", topic="magic")  # type: ignore

    assert output.metadata == mock_get_metadata.return_value
    assert output.response == mock_create.return_value
    assert output.tool_types == mock_tool_types
    assert output.prompt_template == mock_prompt_template
    fn_args = {"genre": "fantasy", "topic": "magic"}
    assert output.fn_args == fn_args
    assert output.messages == mock_messages
    assert output.call_params == mock_create_decorator_kwargs["call_params"]
    assert output.call_kwargs == mock_call_kwargs
    assert (
        output.user_message_param == mock_get_possible_user_message_param.return_value
    )
    assert output.start_time is not None
    assert output.end_time is not None
    assert output._model == "model"

    mock_setup_call_async.assert_called_once_with(
        model="model",
        client=mock_create_decorator_kwargs["client"],
        fn=fn,
        fn_args=fn_args,
        dynamic_config=dynamic_config,
        tools=mock_create_decorator_kwargs["tools"],
        json_mode=mock_create_decorator_kwargs["json_mode"],
        call_params=mock_create_decorator_kwargs["call_params"],
        extract=False,
    )
    mock_create.assert_called_once_with(stream=False, **mock_call_kwargs)



from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class SummaryFeedback(BaseModel):
    """Feedback on summary with a critique and review rewrite based on said critique."""

    critique: str = Field(..., description="The critique of the summary.")
    rewritten_summary: str = Field(
        ...,
        description="A rewritten summary that takes the critique into account.",
    )


@openai.call(model="gpt-4o")
@prompt_template("Summarize the following text into one sentence: {original_text}")
def summarizer(original_text: str): ...


@openai.call(model="gpt-4o", response_model=SummaryFeedback)
@prompt_template(
    """
    Original Text: {original_text}
    Summary: {summary}

    Critique the summary of the original text.
    Then rewrite the summary based on the critique. It must be one sentence.
    """
)
def resummarizer(original_text: str, summary: str): ...


def rewrite_iteratively(original_text: str, summary: str, depth=2):
    text = original_text
    for _ in range(depth):
        text = resummarizer(original_text=text, summary=summary).rewritten_summary
    return text


original_text = """
In the heart of a dense forest, a boy named Timmy pitched his first tent, fumbling with the poles and pegs.
His grandfather, a seasoned camper, guided him patiently, their bond strengthening with each knot tied.
As night fell, they sat by a crackling fire, roasting marshmallows and sharing tales of old adventures.
Timmy marveled at the star-studded sky, feeling a sense of wonder he'd never known.
By morning, the forest had transformed him, instilling a love for the wild that would last a lifetime.
"""

summary = summarizer(original_text=original_text).content
print(f"Summary: {summary}")
# > Summary: In the dense forest, Timmy's first tent-pitching experience with his seasoned camper grandfather deepened their bond and ignited a lifelong love for the wild.
rewritten_summary = rewrite_iteratively(original_text, summary)
print(f"Rewritten Summary: {rewritten_summary}")
# > Rewritten Summary: In the dense forest, Timmy's first tent-pitching experience with his seasoned camper grandfather, filled with roasting marshmallows and sharing tales by the fire, deepened their bond, filled him with wonder at the starry sky, and ignited a lifelong love for the wild.



from pydantic import BaseModel, Field

from mirascope.core import anthropic, prompt_template


class SpellingAndGrammarCheck(BaseModel):
    has_errors: bool = Field(
        description="Whether the text has typos or grammatical errors"
    )


@anthropic.call(
    model="claude-3-5-sonnet-20240620",
    response_model=SpellingAndGrammarCheck,
    json_mode=True,
)
@prompt_template(
    """
    Does the following text have any typos or grammatical errors? {text}
    """
)
def check_for_errors(text: str): ...


text = "Yestday I had a gret time!"
response = check_for_errors(text)
assert response.has_errors



from langfuse.openai import OpenAI

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini", client=OpenAI())
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from pydantic import BaseModel, Field

from mirascope.core.base.call_response import BaseCallResponse
from mirascope.core.base.metadata import Metadata
from mirascope.core.base.stream import BaseStream
from mirascope.core.base.structured_stream import BaseStructuredStream
from mirascope.core.base.tool import BaseTool
from mirascope.integrations.logfire import _utils


class FormatBook(BaseTool):
    """Returns the title and author of a book nicely formatted."""

    title: str = Field(..., description="The title of the book.")
    author: str = Field(..., description="The author of the book in all caps.")

    def call(self) -> str:
        return f"{self.title} by {self.author}"  # pragma: no cover


class MyCallResponse(BaseCallResponse):
    @property
    def content(self) -> str:
        return "content"  # pragma: no cover

    @property
    def tools(self) -> list[BaseTool]:
        return [FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")]


patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
patch.multiple(BaseStream, __abstractmethods__=set()).start()


@patch("logfire.with_settings", new_callable=MagicMock)
def test_logfire_custom_context_manager(mock_logfire: MagicMock) -> None:
    """Tests the `custom_context_manager` context manager."""
    mock_fn = MagicMock()
    mock_fn.__name__ = "mock_fn"
    mock_fn._metadata = Metadata(tags={"tag1", "tag2"})

    with _utils.custom_context_manager(mock_fn):
        mock_logfire.assert_called_once()
        call_args = mock_logfire.call_args[1]
        assert call_args["custom_scope_suffix"] == "mirascope"
        assert isinstance(call_args["tags"], list)
        assert set(call_args["tags"]) == {"tag1", "tag2"}


def test_get_call_response_span_data() -> None:
    call_response = MagicMock()
    result = _utils.get_call_response_span_data(call_response)
    assert result["async"] is False
    assert result["call_params"] == call_response.call_params
    assert result["call_kwargs"] == call_response.call_kwargs
    assert result["model"] == call_response.model
    assert result["provider"] == call_response._provider
    assert result["prompt_template"] == call_response.prompt_template
    assert result["template_variables"] == call_response.fn_args
    assert result["messages"] == call_response.messages
    assert result["response_data"] == call_response.response
    assert result["output"] == {
        "cost": call_response.cost,
        "input_tokens": call_response.input_tokens,
        "output_tokens": call_response.output_tokens,
        "content": call_response.content,
    }


def test_get_tool_calls() -> None:
    call_response = MyCallResponse(
        metadata={"tags": {"version:0001"}},
        response="test response",
        tool_types=[],
        prompt_template="test prompt",
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param={},
        start_time=100,
        end_time=200,
    )  # type: ignore
    result = _utils.get_tool_calls(call_response)
    assert result == [
        {
            "function": {
                "arguments": {
                    "title": "The Name of the Wind",
                    "author": "Rothfuss, Patrick",
                },
                "name": "FormatBook",
            }
        }
    ]

    result = MagicMock()
    result.tools = None
    assert _utils.get_tool_calls(result) is None


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
@patch("mirascope.integrations.logfire._utils.get_tool_calls", new_callable=MagicMock)
def test_handle_call_response(
    mock_get_tool_calls: MagicMock,
    mock_get_call_response_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert _utils.handle_call_response(MagicMock(), mock_fn, None) is None

    result = MagicMock()
    result.tools = [
        FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")
    ]
    span = MagicMock()
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    mock_get_call_response_span_data.return_value = {"output": {}}
    _utils.handle_call_response(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_tool_calls.assert_called_once_with(result)
    mock_get_call_response_span_data.assert_called_once_with(result)
    assert mock_get_call_response_span_data.return_value["async"] is False
    assert (
        set_attributes.call_args[0][0] == mock_get_call_response_span_data.return_value
    )


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
@patch("mirascope.integrations.logfire._utils.get_tool_calls", new_callable=MagicMock)
@pytest.mark.asyncio
async def test_handle_call_response_async(
    mock_get_tool_calls: MagicMock,
    mock_get_call_response_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert await _utils.handle_call_response_async(MagicMock(), mock_fn, None) is None

    result = MagicMock()
    result.tools = [
        FormatBook(title="The Name of the Wind", author="Rothfuss, Patrick")
    ]
    span = MagicMock()
    set_attributes = MagicMock()
    span.set_attributes = set_attributes
    mock_get_call_response_span_data.return_value = {"output": {}}
    await _utils.handle_call_response_async(result, mock_fn, span)
    assert set_attributes.call_count == 1
    mock_get_tool_calls.assert_called_once_with(result)
    mock_get_call_response_span_data.assert_called_once_with(result)
    assert mock_get_call_response_span_data.return_value["async"] is True
    assert (
        set_attributes.call_args[0][0] == mock_get_call_response_span_data.return_value
    )


@patch(
    "mirascope.integrations.logfire._utils.handle_call_response", new_callable=MagicMock
)
def test_handle_stream(mock_handle_call_response: MagicMock) -> None:
    mock_fn = MagicMock()
    mock_stream = MagicMock(spec=BaseStream)
    construct_call_response = MagicMock()
    mock_span = MagicMock()
    mock_stream.construct_call_response.return_value = construct_call_response
    _utils.handle_stream(mock_stream, mock_fn, mock_span)
    mock_handle_call_response.assert_called_once_with(
        construct_call_response, mock_fn, mock_span
    )


@patch(
    "mirascope.integrations.logfire._utils.handle_call_response_async",
    new_callable=AsyncMock,
)
@pytest.mark.asyncio
async def test_handle_stream_async(mock_handle_call_response_async: MagicMock) -> None:
    mock_fn = MagicMock()
    mock_stream = MagicMock(spec=BaseStream)
    construct_call_response = MagicMock()
    mock_span = MagicMock()
    mock_stream.construct_call_response.return_value = construct_call_response
    await _utils.handle_stream_async(mock_stream, mock_fn, mock_span)
    mock_handle_call_response_async.assert_called_once_with(
        construct_call_response, mock_fn, mock_span
    )


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.logfire._utils.set_response_model_output",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_response_model(
    mock_set_response_model_output: MagicMock,
    mock_get_call_response_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert _utils.handle_response_model(MagicMock(), mock_fn, None) is None

    base_model_result = MagicMock(spec=BaseModel)
    base_model_result._response = MagicMock(spec=BaseCallResponse)
    span = MagicMock()
    mock_output = MagicMock()
    mock_set_attributes = MagicMock()
    span.set_attributes = mock_set_attributes
    mock_get_call_response_span_data.return_value = {"output": mock_output}
    _utils.handle_response_model(base_model_result, mock_fn, span)
    mock_get_call_response_span_data.assert_called_once_with(
        base_model_result._response
    )
    mock_set_response_model_output.assert_called_once_with(
        base_model_result, mock_get_call_response_span_data.return_value["output"]
    )
    mock_set_attributes.assert_called_once_with({"output": mock_output, "async": False})


@patch(
    "mirascope.integrations.logfire._utils.get_structured_stream_span_data",
    new_callable=MagicMock,
    return_value={},
)
def test_handle_structured_stream(
    mock_get_structured_stream_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert _utils.handle_structured_stream(MagicMock(), mock_fn, None) is None

    span = MagicMock()

    base_structured_stream_result = MagicMock(spec=BaseStructuredStream)
    base_structured_stream_result.constructed_response_model = MagicMock(spec=BaseModel)
    base_structured_stream_result.stream = MagicMock(spec=BaseStream)
    base_call_response = MagicMock(spec=BaseCallResponse)
    base_structured_stream_result.stream.construct_call_response.return_value = (
        base_call_response
    )
    _utils.handle_structured_stream(base_structured_stream_result, mock_fn, span)
    mock_get_structured_stream_span_data.assert_called_once_with(
        base_structured_stream_result
    )
    assert mock_get_structured_stream_span_data.return_value["async"] is False


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
@patch(
    "mirascope.integrations.logfire._utils.set_response_model_output",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_response_model_async(
    mock_set_response_model_output: MagicMock,
    mock_get_call_response_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert await _utils.handle_response_model_async(MagicMock(), mock_fn, None) is None

    base_model_result = MagicMock(spec=BaseModel)
    base_model_result._response = MagicMock(spec=BaseCallResponse)
    span = MagicMock()
    mock_set_attributes = MagicMock()
    span.set_attributes = mock_set_attributes
    mock_output = MagicMock()
    mock_get_call_response_span_data.return_value = {"output": mock_output}
    await _utils.handle_response_model_async(base_model_result, mock_fn, span)
    mock_get_call_response_span_data.assert_called_once_with(
        base_model_result._response
    )
    mock_set_response_model_output.assert_called_once_with(
        base_model_result, mock_get_call_response_span_data.return_value["output"]
    )
    mock_set_attributes.assert_called_once_with({"output": mock_output, "async": True})


@patch(
    "mirascope.integrations.logfire._utils.get_structured_stream_span_data",
    new_callable=MagicMock,
    return_value={},
)
@pytest.mark.asyncio
async def test_handle_structured_stream_async(
    mock_get_structured_stream_span_data: MagicMock,
) -> None:
    mock_fn = MagicMock()
    assert (
        await _utils.handle_structured_stream_async(MagicMock(), mock_fn, None) is None
    )

    span = MagicMock()

    base_structured_stream_result = MagicMock(spec=BaseStructuredStream)
    base_structured_stream_result.constructed_response_model = MagicMock(spec=BaseModel)
    base_structured_stream_result.stream = MagicMock(spec=BaseStream)
    base_call_response = MagicMock(spec=BaseCallResponse)
    base_structured_stream_result.stream.construct_call_response.return_value = (
        base_call_response
    )
    await _utils.handle_structured_stream_async(
        base_structured_stream_result, mock_fn, span
    )
    mock_get_structured_stream_span_data.assert_called_once_with(
        base_structured_stream_result
    )
    assert mock_get_structured_stream_span_data.return_value["async"] is True


def test_set_response_model_output() -> None:
    class MyBaseModel(BaseModel):
        foo: str

    my_base_model_output = {}
    _utils.set_response_model_output(MyBaseModel(foo="bar"), my_base_model_output)
    assert my_base_model_output == {
        "response_model": {"name": "MyBaseModel", "arguments": {"foo": "bar"}}
    }
    my_base_type_output = {}
    _utils.set_response_model_output("foo", my_base_type_output)
    assert my_base_type_output == {"content": "foo"}


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
def test_get_structured_stream_span_data_base_model(
    mock_get_call_response_span_data: MagicMock,
) -> None:
    class MyBaseModel(BaseModel):
        foo: str

    mock_result = MagicMock(spec=BaseStructuredStream)
    mock_result.stream = MagicMock()
    mock_result.constructed_response_model = MyBaseModel(foo="bar")
    span_data = _utils.get_structured_stream_span_data(mock_result)
    assert span_data == {
        "output": {
            "response_model": {"name": "MyBaseModel", "arguments": {"foo": "bar"}}
        }
    }


@patch(
    "mirascope.integrations.logfire._utils.get_call_response_span_data",
    new_callable=MagicMock,
    return_value={},
)
def test_get_structured_stream_span_data_base_type(
    mock_get_call_response_span_data: MagicMock,
) -> None:
    class MyBaseModel(BaseModel):
        foo: str

    mock_result = MagicMock(spec=BaseStructuredStream)
    mock_result.stream = MagicMock()
    mock_result.constructed_response_model = "foo"
    span_data = _utils.get_structured_stream_span_data(mock_result)
    assert span_data == {"output": {"content": "foo"}}



from openai import OpenAI

from mirascope.core import openai, prompt_template

# Create a custom OpenAI client
custom_client = OpenAI(
    api_key="your-api-key",
    organization="your-organization-id",
    base_url="https://your-custom-endpoint.com/v1",
)


@openai.call(model="gpt-4o-mini", client=custom_client)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



"""Tests the `groq._utils.calculate_cost` function."""

from mirascope.core.groq._utils._calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert (
        calculate_cost(None, None, model="llama3-groq-70b-8192-tool-use-preview")
        is None
    )
    assert calculate_cost(1, 1, model="unknown") is None
    assert (
        calculate_cost(1, 1, model="llama3-groq-70b-8192-tool-use-preview") == 1.78e-6
    )



import asyncio
import random

from mirascope.core import openai, prompt_template

qa_examples = [
    "Q: What are your company's core values?\nA: Our company's core values are integrity, innovation, customer-centricity, and teamwork. We believe that maintaining these values is crucial to achieving our mission and vision.",
    "Q: How do you handle conflicts in the workplace?\nA: We handle conflicts by promoting open communication, understanding different perspectives, and seeking mutually beneficial solutions. We have clear policies and trained mediators to assist in resolving conflicts effectively.",
    "Q: Can you describe a time when you exceeded a client's expectations?\nA: Certainly. Recently, we completed a project ahead of schedule and under budget. We also provided additional insights and recommendations that significantly benefited the client, earning their gratitude and loyalty.",
    "Q: How do you ensure continuous improvement within the team?\nA: We ensure continuous improvement by encouraging regular training, fostering a culture of feedback, and implementing agile methodologies. We also review our processes regularly to identify areas for enhancement.",
    "Q: What strategies do you use to stay ahead of industry trends?\nA: We stay ahead of industry trends by investing in research and development, attending industry conferences, and maintaining strong relationships with thought leaders. We also encourage our team to engage in continuous learning and innovation.",
    "Q: How do you measure the success of a project?\nA: We measure the success of a project by evaluating key performance indicators such as client satisfaction, budget adherence, timeline compliance, and the quality of the deliverables. Post-project reviews help us to identify successes and areas for improvement.",
    "Q: What is your approach to diversity and inclusion?\nA: Our approach to diversity and inclusion involves creating a welcoming environment for all employees, offering diversity training, and implementing policies that promote equality. We value diverse perspectives as they drive innovation and growth.",
    "Q: How do you manage remote teams effectively?\nA: We manage remote teams effectively by leveraging technology for communication and collaboration, setting clear goals, and maintaining regular check-ins. We also ensure that remote employees feel included and supported.",
    "Q: What are your short-term and long-term goals for the company?\nA: In the short term, our goals include expanding our market reach and enhancing our product offerings. In the long term, we aim to become industry leaders by driving innovation and achieving sustainable growth.",
    "Q: How do you handle feedback from clients or customers?\nA: We handle feedback by listening actively, responding promptly, and taking necessary actions to address concerns. We view feedback as an opportunity for improvement and strive to exceed our clients' expectations continuously.",
]


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Here are some examples that demonstrate the voice to use in a corporate setting.
    {examples}

    With these examples, answer the following question:
    {query}
    """
)
async def answer(query: str) -> openai.OpenAIDynamicConfig:
    random_indices = random.sample(range(len(qa_examples)), 3)
    examples = [qa_examples[i] for i in random_indices]
    return {"computed_fields": {"examples": examples}}


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Take the following responses from an LLM and aggregate/average them into
    one answer.
    {responses}
    """
)
async def aggregate_answers(
    query: str, num_responses: int
) -> openai.OpenAIDynamicConfig:
    tasks = [answer(query) for _ in range(num_responses)]
    responses = await asyncio.gather(*tasks)
    # Uncomment to see intermediate responses
    print([response.content for response in responses])

    return {
        "computed_fields": {"responses": [response.content for response in responses]}
    }


query = """Help me write a notice that highlights the importance of attending\
the company's social events. Give me just the notice, no further explanation."""


async def demonstration_ensembling(query, ensemble_size):
    print(await aggregate_answers(query=query, num_responses=ensemble_size))


asyncio.run(demonstration_ensembling(query=query, ensemble_size=5))
# > **Notice: Importance of Attending Company Social Events**
#
#   Dear Team,
#
#   We want to highlight the significance of participating in our upcoming company social events. These gatherings are valuable opportunities for team bonding, networking, and fostering a positive workplace culture. Engaging with colleagues outside of our regular work environment enhances collaboration, strengthens relationships, and supports a cohesive team atmosphere.
#
#   Your presence at these events not only enriches the experience for yourself but also contributes greatly to our vibrant community. We encourage everyone to attend, connect with one another, and celebrate our successes together.
#
#   Thank you for your commitment to making our workplace a supportive and dynamic environment.
#
#   Best regards,
#   [Your Name]
#   [Your Position]


# Intermediate Responses

# print([response.content for response in responses])
# NOTICE: Importance of Attending Company Social Events

# Dear Team,

# We are excited to remind everyone about the significance of participating in our upcoming social events. These gatherings provide a unique opportunity to foster connections, strengthen team dynamics, and enhance our company culture. Engaging in social activities allows us to build relationships outside of our typical work environment, promoting collaboration and teamwork.

# Your presence is vital in creating a vibrant and inclusive atmosphere where everyone feels valued and connected. Let’s come together to celebrate our successes and continue to build a strong community within our organization.

# We look forward to seeing you all there!

# Best Regards,
# [Your Name]
# [Your Position]

# Notice: Importance of Attending Company Social Events

# Dear Team,

# We would like to emphasize the significance of participating in our company social events. These gatherings provide valuable opportunities for team building, strengthening relationships, and fostering a positive workplace culture. Engaging with colleagues outside of our regular work routine enhances collaboration and communication, ultimately contributing to our shared success.

# We encourage everyone to make an effort to attend these events. Your participation not only enriches the experience for yourself but also supports a cohesive and motivated team environment.

# Thank you for your commitment to making our workplace a vibrant community.

# Best regards,
# [Your Company Name]

# Notice: Importance of Attending Company Social Events

# Dear Team,

# We would like to emphasize the importance of participating in our upcoming company social events. These gatherings provide a valuable opportunity for team bonding, networking, and fostering a positive workplace culture. Engaging with your colleagues outside of the usual work environment helps strengthen relationships, promotes collaboration, and enhances overall team spirit.

# We encourage everyone to attend and make the most of these events. Your presence not only enriches the experience but also contributes to the vibrant culture we strive to create at our company.

# Thank you for your participation and enthusiasm!

# Best regards,
# [Your Name]
# [Your Position]

# Notice: Importance of Attending Company Social Events

# Dear Team,

# We would like to emphasize the significance of participating in our upcoming social events. These gatherings provide an excellent opportunity for team building, fostering relationships, and enhancing our corporate culture. Connecting with colleagues outside of the traditional work environment allows for stronger collaboration and communication across departments.

# Your presence not only contributes to a positive workplace atmosphere but also helps in cultivating a sense of community within our organization. We encourage everyone to engage, share ideas, and connect with one another during these events.

# Thank you for your participation!

# Best regards,
# [Your Name]
# [Your Position]

# Notice: Importance of Attending Company Social Events

# Dear Team,

# We would like to take a moment to emphasize the significance of participating in our upcoming company social events. These gatherings are not only an opportunity to unwind and enjoy each other's company, but they also play a crucial role in fostering teamwork, collaboration, and a sense of community within our organization.

# Attending these events allows you to connect with colleagues from different departments, share experiences, and build relationships that enhance our workplace culture. It is through these interactions that we can strengthen our bonds and create a more cohesive and productive environment.

# We encourage everyone to make time for these events, as your participation is vital to the success of our team dynamics and overall morale. Let’s come together to celebrate our achievements, support one another, and continue to promote a positive workplace atmosphere.

# Thank you for your commitment to making our company a great place to work.

# Best regards,
# [Your Name]
# [Your Title]



import asyncio

from pydantic import BaseModel

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Please identify a chef who is well known for cooking with {ingredient}.
    Respond only with the chef's name.
    """
)
async def chef_selector(ingredient: str): ...


class IngredientsList(BaseModel):
    ingredients: list[str]


@openai.call(model="gpt-4o-mini", response_model=IngredientsList)
@prompt_template(
    """
    Given a base ingredient {ingredient}, return a list of complementary ingredients.
    Make sure to exclude the original ingredient from the list.
    """
)
async def ingredients_identifier(ingredient: str): ...


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    SYSTEM:
    Your task is to recommend a recipe. Pretend that you are chef {chef}.

    USER:
    Recommend recipes that use the following ingredients:
    {ingredients}
    """
)
async def recipe_recommender(ingredient: str) -> openai.OpenAIDynamicConfig:
    chef, ingredients = await asyncio.gather(
        chef_selector(ingredient), ingredients_identifier(ingredient)
    )
    return {"computed_fields": {"chef": chef, "ingredients": ingredients}}


async def run():
    response = await recipe_recommender(ingredient="apples")
    print(response.content)


asyncio.run(run())



import re

from llama_index.core import (
    Settings,
    SimpleDirectoryReader,
    VectorStoreIndex,
    load_index_from_storage,
)
from llama_index.core.base.response.schema import Response
from llama_index.core.postprocessor import LLMRerank
from llama_index.core.storage import StorageContext
from llama_index.core.vector_stores import SimpleVectorStore
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.llms.ollama import Ollama
from openai import OpenAI
from pydantic import BaseModel

from mirascope.core import openai, prompt_template

Settings.llm = Ollama(model="llama3.1")
Settings.embed_model = HuggingFaceEmbedding(model_name="BAAI/bge-small-en-v1.5")

# ONE TIME SETUP
documents = SimpleDirectoryReader("PATH/TO/YOUR/DOCS").load_data()
vector_store = SimpleVectorStore()
storage_context = StorageContext.from_defaults(vector_store=vector_store)
index = VectorStoreIndex.from_documents(documents, storage_context=storage_context)
index.storage_context.persist()
# END ONE TIME SETUP

storage_context = StorageContext.from_defaults(persist_dir="storage")
loaded_index = load_index_from_storage(storage_context)
query_engine = loaded_index.as_query_engine()


def custom_parse_choice_select_answer_fn(
    answer: str, num_choices: int, raise_error: bool = False
) -> tuple[list[int], list[float]]:
    """Custom parse choice select answer function."""
    answer_lines = answer.split("\n")
    answer_nums = []
    answer_relevances = []
    for answer_line in answer_lines:
        line_tokens = answer_line.split(",")
        if len(line_tokens) != 2:
            if not raise_error:
                continue
            else:
                raise ValueError(
                    f"Invalid answer line: {answer_line}. "
                    "Answer line must be of the form: "
                    "answer_num: <int>, answer_relevance: <float>"
                )
        split_tokens = line_tokens[0].split(":")
        if (
            len(split_tokens) != 2
            or split_tokens[1] is None
            or not split_tokens[1].strip().isdigit()
        ):
            continue
        answer_num = int(line_tokens[0].split(":")[1].strip())
        if answer_num > num_choices:
            continue
        answer_nums.append(answer_num)
        # extract just the first digits after the colon.
        _answer_relevance = re.findall(r"\d+", line_tokens[1].split(":")[1].strip())[0]
        answer_relevances.append(float(_answer_relevance))
    return answer_nums, answer_relevances


def get_documents(query: str) -> str:
    """The get_documents tool that retrieves Mirascope documentation based on the
    relevance of the query"""
    query_engine = loaded_index.as_query_engine(
        similarity_top_k=10,
        node_postprocessors=[
            LLMRerank(
                choice_batch_size=5,
                top_n=2,
                parse_choice_select_answer_fn=custom_parse_choice_select_answer_fn,
            )
        ],
        response_mode="tree_summarize",
    )

    response = query_engine.query(query)
    if isinstance(response, Response):
        return response.response or "No documents found."
    return "No documents found."


class MirascopeBot(BaseModel):
    @openai.call(
        model="llama3.1",
        client=OpenAI(base_url="http://localhost:11434/v1", api_key="ollama"),
    )
    @prompt_template(
        """
        SYSTEM:
        You are an AI Assistant that is an expert at answering questions about Mirascope.
        Here is the relevant documentation to answer the question.

        Context:
        {context}

        USER:
        {question}
        """
    )
    def _step(self, context: str, question: str): ...

    def _get_response(self, question: str):
        context = get_documents(question)
        answer = self._step(context, question)
        print("(Assistant):", answer.content)
        return

    def run(self):
        while True:
            question = input("(User): ")
            if question == "exit":
                break
            self._get_response(question)


MirascopeBot().run()



from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel

from mirascope.core import openai, prompt_template


class Book(BaseModel):
    title: str
    author: str


class Librarian(BaseModel):
    history: list[ChatCompletionMessageParam] = []
    books: dict[str, str] = {}

    def add_book(self, book: Book) -> str:
        """Adds a book to the collection."""
        self.books[book.title] = book.author
        return f"Added '{book.title}' by {book.author} to the collection."

    def remove_book(self, book: Book) -> str:
        """Removes a book from the collection."""
        self.books.pop(book.title, None)
        return f"Removed '{book.title}' by {book.author} from the collection."

    @openai.call(model="gpt-4o-mini", stream=True)
    @prompt_template(
        """
        SYSTEM:
        You are a helpful librarian assistant. You have access to the following tools:
        - `add_book`: Adds a book to your collection of books.
        - `remove_book`: Removes a book from your collection of books.
        You should use these tools to update your book collection according to the user.
        For example, if the user mentions that they recently read a book that they liked, consider adding it to the collection.
        If the user didn't like a book, consider removing it from the collection.

        You currently have the following books in your collection:
        {self.books}

        MESSAGES: {self.history}
        USER: {query}
        """
    )
    def _call(self, query: str) -> openai.OpenAIDynamicConfig:
        return {"tools": [self.add_book, self.remove_book]}

    def _step(self, query: str) -> str:
        response = self._call(query)
        tools_and_outputs = []
        for chunk, tool in response:
            if tool:
                tools_and_outputs.append((tool, tool.call()))
            else:
                print(chunk.content, end="", flush=True)

        if response.user_message_param:
            self.history.append(response.user_message_param)
        self.history.append(response.message_param)
        if tools_and_outputs:
            self.history += response.tool_message_params(tools_and_outputs)
            return self._step("")
        else:
            return response.content

    def run(self):
        while True:
            query = input("User: ")
            if query.lower() == "exit":
                break

            print("Librarian: ", end="", flush=True)
            self._step(query)
            print("")


librarian = Librarian()
librarian.run()



from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    Answer this question, thinking step by step.
    {query}
    """
)
def cot_call(query: str): ...


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    {query}
    It's very important to my career.
    """
)
def emotion_prompting_call(query: str): ...


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    {query}
    Rephrase and expand the question, and respond.
    """
)
def rar_call(query: str): ...


class BestResponse(BaseModel):
    best_response: str = Field(
        ..., description="The best response of the options given, verbatim"
    )
    reasoning: str = Field(
        ...,
        description="""A short description of why this is the best response to
        the query, along with reasons why the other answers were worse.""",
    )


@openai.call(model="gpt-4o-mini", response_model=BestResponse)
@prompt_template(
    """
    Here is a query: {query}

    Evaluate the following responses from LLMs and decide which one
    is the best based on correctness, fulfillment of the query, and clarity:

    Response 1:
    {cot_response}

    Response 2:
    {emotion_prompting_response}

    Response 3:
    {rar_response}
    """
)
def mixture_of_reasoning(query: str) -> openai.OpenAIDynamicConfig:
    cot_response = cot_call(query=query)
    emotion_prompting_response = emotion_prompting_call(query=query)
    rar_response = rar_call(query=query)
    # Uncomment to see intermediate responses
    # print(cot_response)
    # print(emotion_prompting_response)
    # print(rar_response)

    return {
        "computed_fields": {
            "cot_response": cot_response,
            "emotion_prompting_response": emotion_prompting_response,
            "rar_response": rar_response,
        }
    }


prompt = """What are the side lengths of a rectangle with area 8 and perimeter 12?"""

best_response = mixture_of_reasoning(prompt)
print(best_response.best_response)
# > Response 2: The best response is Response 2. It directly and correctly solves the system of equations to find the side lengths of the rectangle with the area 8 and perimeter 12. It provides clear steps and demonstrates the correct approach to the problem, yielding the correct result of 2 units by 4 units.

print(best_response.reasoning)
# > Response 1 overcomplicates the solution process by introducing unnecessary steps and expressions. Response 3 contains an error in the calculation and solution steps, providing incorrect side lengths for the rectangle. Response 2 simplifies the equations and solves for the side lengths accurately and concisely, making it the most effective response.

# Intermediate Responses

# cot_response = cot_call(query=query)
# To find the side lengths of a rectangle with area 8 and perimeter 12, we need to denote the length and width of the rectangle as variables. Let's use L for the length and W for the width.

# Step 1: Identify the formulas for the area and perimeter of a rectangle.
# Area of a rectangle = length x width
# Perimeter of a rectangle = 2(length + width)

# Step 2: Given that the area of the rectangle is 8, set up the equation using the formula for the area.
# L x W = 8

# Step 3: Given that the perimeter of the rectangle is 12, set up the equation using the formula for the perimeter.
# 2(L + W) = 12
# L + W = 6

# Step 4: We now have a system of equations:
# L x W = 8
# L + W = 6

# Step 5: Solve the system of equations simultaneously to find the values of L and W. Let's substitute one of the variables to solve for the other:
# L = 6 - W

# Step 6: Substitute L = 6 - W into the equation L x W = 8:
# (6 - W) x W = 8
# 6W - W^2 = 8
# W^2 - 6W + 8 = 0

# Step 7: Solve the quadratic equation for W. Factoring or using the quadratic formula will give you the values of W.

# Step 8: Once you have the value(s) of W, substitute back into L = 6 - W to find the corresponding value(s) of L.

# Step 9: Check the solutions to ensure they satisfy the conditions given in the question.

# By following these steps, you can determine the side lengths of the rectangle with area 8 and perimeter 12.


# emotion_prompting_response = emotion_prompting_call(query=query)
# Let's denote the length of the rectangle as L and the width as W.

# The area of a rectangle is given by the formula Area = Length x Width. So in this case, we have that LW = 8.

# The perimeter of a rectangle is given by the formula Perimeter = 2(Length + Width). So in this case, we have that 2(L + W) = 12.

# We are looking to solve the system of equations:

# LW = 8
# 2(L + W) = 12

# From the first equation, we can rewrite it as W = 8/L.

# Substitute this into the second equation:

# 2(L + 8/L) = 12
# 2L + 16/L = 12
# 2L^2 + 16 = 12L
# 2L^2 - 12L + 16 = 0

# Solving this quadratic equation, we get L = 2 or L = 8.

# Therefore, the possible sides of the rectangle are 2 and 4 or 4 and 2.


# rar_response = rar_call(query=query)
# What are the dimensions of a rectangle with an area of 8 square units and a perimeter of 12 units?

# Let's denote the length of the rectangle as L units and the width as W units.
# Given that the area of the rectangle is 8 units², we have the equation:
# L x W = 8

# Also, the perimeter of a rectangle is given by the formula:
# 2L + 2W = 12

# We are asked to find the values of L and W that satisfy both of these equations. Solving these two equations simultaneously, we can find the side lengths of the rectangle. Let's start by substituting L = 8/W into the perimeter equation:

# 2(8/W) + 2W = 12
# 16/W + 2W = 12
# 16 + 2W² = 12W
# 2W² -12W + 16 = 0
# W² - 6W + 8 = 0
# (W-4)(W-2) = 0
# W=4 or W=2

# Therefore, the possible values for W are 4 units or 2 units.
# Substitute W = 4 into L x W = 8:
# L x 4 = 8
# L = 2 units
# So, the dimensions of the rectangle are 2 units by 4 units.



"""This module contains the `GroqCallResponseChunk` class.

usage docs: learn/streams.md#handling-streamed-responses
"""

from groq.types.chat import ChatCompletionChunk
from groq.types.chat.chat_completion import Choice
from groq.types.completion_usage import CompletionUsage

from ..base import BaseCallResponseChunk

FinishReason = Choice.__annotations__["finish_reason"]


class GroqCallResponseChunk(BaseCallResponseChunk[ChatCompletionChunk, FinishReason]):
    """A convenience wrapper around the Groq `ChatCompletionChunk` streamed chunks.

    When calling the Groq API using a function decorated with `groq_call` and
    `stream` set to `True`, the stream will contain `GroqResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.groq import groq_call


    @groq_call("llama-3.1-8b-instant", stream=True)
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    stream = recommend_book("fantasy")  # response is an `GroqStream`
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        delta = None
        if self.chunk.choices:
            delta = self.chunk.choices[0].delta
        return delta.content if delta is not None and delta.content else ""

    @property
    def finish_reasons(
        self,
    ) -> list[FinishReason]:
        """Returns the finish reasons of the response."""
        return [
            choice.finish_reason
            for choice in self.chunk.choices
            if choice.finish_reason
        ]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.chunk.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.chunk.id

    @property
    def usage(self) -> CompletionUsage | None:
        """Returns the usage of the chat completion."""
        if self.chunk.usage:
            return self.chunk.usage
        return None

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.prompt_tokens
        return None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.completion_tokens
        return None



"""Tests the `_utils.get_possible_user_message_param` module."""

from pydantic import BaseModel

from mirascope.core.base._utils._get_possible_user_message_param import (
    get_possible_user_message_param,
)


def test_get_possible_user_message_param() -> None:
    """Tests the `get_possible_user_message_param` function."""
    assert get_possible_user_message_param([]) is None

    messages = [{"role": "user", "text": "Hello, world!"}]
    assert get_possible_user_message_param(messages) == messages[0]

    messages = [{"role": "assistant", "text": "Hi!"}]
    assert get_possible_user_message_param(messages) is None

    messages = [
        {"role": "user", "text": "Hello, world!"},
        {"role": "assistant", "text": "Hi!"},
    ]
    assert get_possible_user_message_param(messages) is None

    messages = [
        {"role": "user", "text": "Hello!"},
        {"role": "assistant", "text": "Hi!"},
        {"role": "user", "text": "How are you?"},
    ]
    assert get_possible_user_message_param(messages) == messages[-1]

    class Message(BaseModel):
        role: str
        text: str

    messages = [Message(role="user", text="Hello, world!")]
    assert get_possible_user_message_param(messages) == messages[0]

    messages = [Message(role="assistant", text="Hi!")]
    assert get_possible_user_message_param(messages) is None

    messages = [
        Message(role="user", text="Hello, world!"),
        Message(role="assistant", text="Hi!"),
    ]
    assert get_possible_user_message_param(messages) is None

    messages = [
        Message(role="user", text="Hello!"),
        Message(role="assistant", text="Hi!"),
        Message(role="user", text="How are you?"),
    ]
    assert get_possible_user_message_param(messages) == messages[-1]



"""Tests the `mistral._utils.convert_message_params` function."""

import pytest
from mistralai.models.chat_completion import ChatMessage

from mirascope.core.base import AudioPart, BaseMessageParam, ImagePart, TextPart
from mirascope.core.mistral._utils._convert_message_params import convert_message_params


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[BaseMessageParam | ChatMessage] = [
        ChatMessage(role="user", content="Hello"),
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(role="user", content=[TextPart(type="text", text="Hello")]),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        ChatMessage(role="user", content="Hello"),
        ChatMessage(role="user", content="Hello"),
        ChatMessage(role="user", content="Hello"),
    ]

    with pytest.raises(
        ValueError,
        match="Mistral does not currently support multimodalities.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/jpeg",
                            image=b"image",
                            detail="auto",
                        )
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="Mistral does not currently support multimodalities.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio",
                            media_type="audio/wav",
                            audio=b"audio",
                        )
                    ],
                ),
            ]
        )



from enum import Enum

from mirascope.core import openai, prompt_template


class Sentiment(Enum):
    NEGATIVE = "negative"
    NEUTRAL = "neutral"
    POSITIVE = "positive"


@openai.call("gpt-4o-mini", response_model=Sentiment)
@prompt_template(
    """
    Classify the sentiment of the following text: {text}
    """
)
def classify_sentiment(text: str): ...


text = "I hate this product. It's terrible."
label = classify_sentiment(text)
assert label == Sentiment.NEGATIVE

text = "I don't feel strongly about this product."
label = classify_sentiment(text)
assert label == Sentiment.NEUTRAL

text = "I love this product. It's amazing!"
label = classify_sentiment(text)
assert label == Sentiment.POSITIVE



"""This module defines the function return type for functions as LLM calls.

usage docs: learn/dynamic_configuration.md#dynamic-configuration-options
"""

from google.generativeai.types import ContentsType  # type: ignore

from ..base import BaseDynamicConfig
from .call_params import GeminiCallParams

GeminiDynamicConfig = BaseDynamicConfig[ContentsType, GeminiCallParams]
"""The function return type for functions wrapped with the `gemini_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.gemini import GeminiDynamicConfig, gemini_call


@gemini_call("gemini-flash-1.5")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> GeminiDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""This module contains the `MistralCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from mistralai.models.chat_completion import ChatCompletionResponse, ChatMessage
from mistralai.models.common import UsageInfo
from pydantic import computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import MistralCallParams
from .dynamic_config import MistralDynamicConfig
from .tool import MistralTool


class MistralCallResponse(
    BaseCallResponse[
        ChatCompletionResponse,
        MistralTool,
        MistralDynamicConfig,
        ChatMessage,
        MistralCallParams,
        ChatMessage,
    ]
):
    """A convenience wrapper around the Mistral `ChatCompletion` response.

    When calling the Mistral API using a function decorated with `mistral_call`, the
    response will be an `MistralCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.mistral import mistral_call


    @mistral_call("mistral-largel-latest")
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")  # response is an `MistralCallResponse` instance
    print(response.content)
    ```
    """

    _provider = "mistral"

    @property
    def content(self) -> str:
        """The content of the chat completion for the 0th choice."""
        return self.response.choices[0].message.content

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [
            choice.finish_reason if choice.finish_reason else ""
            for choice in self.response.choices
        ]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> UsageInfo:
        """Returns the usage of the chat completion."""
        return self.response.usage

    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> ChatMessage:
        """Returns the assistants's response as a message parameter."""
        return self.response.choices[0].message

    @computed_field
    @property
    def tools(self) -> list[MistralTool] | None:
        """Returns the tools for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tool_calls = self.response.choices[0].message.tool_calls
        if not self.tool_types or not tool_calls:
            return None

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> MistralTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[MistralTool, str]]
    ) -> list[ChatMessage]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The list of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `ChatMessage` parameters.
        """
        return [
            ChatMessage(
                role="tool",
                content=output,
                tool_call_id=tool.tool_call.id,
                name=tool._name(),
            )
            for tool, output in tools_and_outputs
        ]



from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, wait_exponential

from mirascope.core import anthropic, prompt_template


class Book(BaseModel):
    title: str
    author: str


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@anthropic.call("claude-3-5-sonnet-20240620", response_model=Book)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


book = recommend_book("fantasy")
print(book)
